"""
This package implements some basic infrastructure for handling data
streams in a structured manner. It contains three distinct classes:

- `BaseMsg` transports Python objects, using `send` and `recv`.
- `BaseBlk` transports delimited bytestrings, using `snd` and `rcv`.
- `BaseBuf` transports undelimited bytestrings, using `wr` and `rd`.

Additionally, message- and block-based classes understand `cwr` and `crd`
which transport out-of-band data. Typically these contain raw console bytes
that are interleaved with structured data; they are used on channels which
needs to multiplex both.

The common theme for using these classes is

- You assemble a communication stack bottom-up: start with a serial link;
  add packetizing, retransmission, and an object-codec.
- Using the top level as an async context managers establishes the complete
  stack; leaving the context tears it down.
- Most likely, connect the result to a [Base]MsgHandler.

Everything is fully asynchronous. There is no "new incoming data" callback:
it's the upper layer's job to repeatedly call the appropriate method to
read or receive new messages.

A `wrap` method provides a secondary context that can be used for
a persistent outer context, e.g. to keep a listening socket open.
"""

from __future__ import annotations

from moat.util import attrdict
from moat.lib.micro import ACM, AC_exit, AC_use

# Typing

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager
    from types import TracebackType

    from typing import Any, Self

    Buffer = bytes | bytearray | memoryview
    MutBuffer = bytearray | memoryview
    _AACMBase = AbstractAsyncContextManager
else:
    Buffer = bytes
    MutBuffer = bytearray
    _AACMBase = object


class _NullCtx:
    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *tb):
        pass


_nullctx = _NullCtx()


class Base(_AACMBase):
    """
    The MoaT stream base class for "something connected".

    This class *must* be used as an async context manager.

    Usage:

    Use the :py:func:`moat.lib.micro.AC_use` helper if you need
    to call an async context manager or to register a destructor.

    Augment `setup` or `teardown` to add non-stream related features.

    Override `wrap` to contain an async context manager that holds resources
    which must survive reconnection, e.g. a MQTT link's persistent state or
    a listening socket.
    """

    s = None
    cfg: attrdict

    def __init__(self, cfg: attrdict | None = None):
        if cfg is None:
            cfg = attrdict()
        self.cfg = cfg

    @property
    def cfg_name(self):
        "Some name to disambiguate this part"
        try:
            return self.cfg.name
        except AttributeError:
            return self.__class__.__name__

    def wrap(self) -> AbstractAsyncContextManager:
        """
        Async context manager for holding a cross-connection context.

        By default does nothing.
        """
        return _nullctx

    async def __aenter__(self) -> Self:
        res = self
        AC = ACM(self)
        await AC(self.teardown)
        try:
            await self.setup()
            if (ctx := getattr(self, "_ctx", None)) is not None:
                res = await AC(ctx())
            return res
        except BaseException as exc:
            await AC_exit(self, type(exc), exc, getattr(exc, "__traceback__", None))
            raise

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> bool | None:
        return await AC_exit(self, exc_type, exc, tb)

    async def setup(self):
        """
        Basic async setup method. You may use AC_exit.

        Call the superclass first, when overriding.
        """

    async def teardown(self):
        """
        Object destructor.

        Should not fail when called with a partially-created object.
        """


class BaseConn(Base):
    """
    This is the MoaT stream base class for "something connected that talks".

    This class *must* be used as an async context manager.

    Usage:

    Override :py:meth:`BaseConn.stream` to create the data link.
    Use :py:func:`moat.lib.micro.AC_use` to
    call an async context manager or to register a destructor.

    Augment `setup` or `teardown` to add non-stream related features.
    """

    s: Any = None

    async def setup(self):
        """
        Object construction.

        By default, assigns the result of calling :py:meth:`BaseConn.stream` to the attribute
        ``s``.
        """
        if self.s is not None:
            raise RuntimeError("Busy!")

        self.s = await self.stream()

    async def teardown(self):
        """
        Object destructor.

        Should not fail when called with a partially-created object.
        """
        self.s = None

    async def stream(self):
        """
        Data stream setup.

        You need to use `moat.lib.micro.AC_use` for setting up an async context
        or to register a cleanup handler.
        """
        raise NotImplementedError(f"'stream' in {self!r}")


class BaseMsg(BaseConn):
    """
    A stream base module for messages. May not be useful.

    Implement send/recv.
    """

    async def send(self, m: Any) -> Any:
        """
        Send a message.
        """
        raise NotImplementedError(f"'send' in {self!r}")

    async def recv(self) -> Any:
        """
        Receive a message.
        """
        raise NotImplementedError(f"'recv' in {self!r}")


class BaseBlk(BaseConn):
    """
    A stream base module for bytestrings. May not be useful.

    Implement snd/rcv.
    """

    async def snd(self, m: Buffer | bytes) -> None:
        """
        Send a block of bytes.
        """
        raise NotImplementedError(f"'send' in {self!r}")

    async def rcv(self) -> Buffer | bytes:
        """
        Receive a block of bytes.
        """
        raise NotImplementedError(f"'recv' in {self!r}")


class BaseBuf(BaseConn):
    """
    A stream base module for bytestreams.

    Implement rd/wr.
    """

    async def rd(self, buf: MutBuffer) -> int:
        """
        Read some bytes.

        @buf is a bytearray to read data into. The return value is the
        number of bytes filled.

        This method never returns zero. End-of-file raises `EOFError`.
        """
        raise NotImplementedError(f"'rd' in {self!r}")

    async def wr(self, data: Buffer) -> int:
        """
        Write some bytes.
        """
        raise NotImplementedError(f"'wr' in {self!r}")


class StackedConn(BaseConn):
    """
    Base class for connection stacking.

    Connection stacks have a lower layer. Our `stream` methods uses it as an
    async context manager to create connection from it, using its

    Args:
        link(BaseConn): The lower layer to run on top of.
        cfg: Config data
    """

    link: BaseConn | None = None

    def __init__(self, link: BaseConn, cfg):
        super().__init__(cfg=cfg)
        self.link = link

    def wrap(self):  # noqa:D102
        if self.link is None:
            raise RuntimeError("No link")
        return self.link.wrap()

    async def stream(self) -> BaseConn:
        """
        Generate the low-level connection this module uses.

        By default, returns the linked stream's async context.
        """
        if self.link is None:
            raise RuntimeError("No link")
        return await AC_use(self, self.link)


class StackedMsg(StackedConn, BaseMsg):
    """
    A no-op stack module for messages. Override to implement interesting features.

    Use the attribute "s" to store the linked stream's context.

    Args:
        link(BaseMsg): The lower layer to run on top of.
        cfg: Config data
    """

    async def send(self, m: Any) -> Any:
        "Send. Transmits a structured message"
        return await self.s.send(m)

    async def recv(self) -> Any:
        "Receive. Returns a message."
        return await self.s.recv()

    async def cwr(self, buf: Buffer) -> None:
        "Console Send. Returns when the buffer is transmitted."
        await self.s.cwr(buf)

    async def crd(self, buf: MutBuffer) -> int:
        "Console Receive. Returns data by reading into a buffer."
        return await self.s.crd(buf)


class StackedBuf(StackedConn, BaseBuf):
    """
    A no-op stack module for byte steams. Override to implement interesting features.

    Use the attribute "s" to store the linked stream's context.

    Args:
        link(BaseBuf): The lower layer to run on top of.
        cfg: Config data
    """

    async def wr(self, data: Buffer) -> int:
        "Send. Returns when the buffer is transmitted."
        return await self.s.wr(data)

    async def rd(self, buf: MutBuffer) -> int:
        "Receive. Returns data by reading into a buffer."
        return await self.s.rd(buf)


class StackedBlk(StackedConn, BaseBlk):
    """
    A no-op stack module for bytestrings. Override to implement interesting features.

    Use the attribute "s" to store the linked stream's context.

    Args:
        link(BaseBlk): The lower layer to run on top of.
        cfg: Config data
    """

    cwr = StackedMsg.cwr
    crd = StackedMsg.crd

    async def snd(self, m: Buffer | bytes) -> None:
        "Send. Transmits a structured message"
        await self.s.snd(m)

    async def rcv(self) -> Buffer | bytes:
        "Receive. Returns a message."
        return await self.s.rcv()



"""
Raw TCP stream app.
"""

from __future__ import annotations

from moat.util import attrdict
from moat.lib.micro import AC_use
from moat.lib.rpc.stream.cmdbbm import BaseCmdBBM
from moat.lib.stream import TcpLink

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from types import CoroutineType

    from moat.lib.stream import BaseBuf

    from typing import Any


class Raw(BaseCmdBBM):
    """Sends/receives raw data over TCP."""

    def stream(self) -> CoroutineType[Any, Any, BaseBuf]:
        """Returns the TCP stream."""
        return AC_use(
            self,
            TcpLink(
                self.cfg.get("host", "127.0.0.1"),
                self.cfg["port"],
                retry=self.cfg.get("retry", attrdict()),
            ),
        )



"""
A hacked-up copy of some parts of `moat.util`.
"""

from __future__ import annotations

import re

_PartRE = re.compile("[^:._]+|_|:|\\.")

b64decode = None


class Path(tuple):  # noqa:SLOT001
    """
    somewhat-dummy Path

    half-assed string analysis, somewhat-broken output for non-basics
    """

    def __str__(self):
        def _escol(x):
            x = x.replace(":", "::").replace(".", ":.").replace(" ", ":_")
            return x

        res = []
        if not len(self):
            res.append(":")
        for x in self:
            if isinstance(x, str):
                if x == "":
                    res.append(":e")
                else:
                    if res:
                        res.append(".")
                    res.append(_escol(x))
            elif x is True:
                res.append(":t")
            elif x is False:
                res.append(":f")
            elif x is None:
                res.append(":n")
            elif isinstance(x, (bytes, bytearray, memoryview)):
                if all(32 <= b < 127 for b in x):
                    res.append(":v" + _escol(x.decode("ascii")))
                else:
                    from base64 import b64encode  # noqa: PLC0415

                    res.append(":s" + b64encode(x).decode("ascii"))
                    # no hex
            else:
                res.append(":" + _escol(repr(x)))
        return "".join(res)

    @classmethod
    def build(cls, arr, decoded=False):  # noqa:ARG003
        """
        Constructor to build a Path from an existing list.
        """
        return cls(arr)

    def __repr__(self):
        return f"P({str(self)!r})"

    def __truediv__(self, x):
        return Path(self + (x,))

    def __add__(self, x):
        return Path(tuple(self) + tuple(x))

    def __radd__(self, x):
        return Path(tuple(x) + tuple(self))



"""
Remote command forwarding app.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from types import CoroutineType

    from moat.lib.path import PathElem
    from moat.lib.rpc import MsgSender
    from moat.lib.rpc.msg import Msg

    from typing import Any


class Fwd(BaseCmd):
    """
    Link to a stream that's someplace else.

    This app forwards to somewhere else.
    """

    sd: MsgSender

    doc = dict(_c=dict(_d="Command forwarding", path="path:dest"))

    async def setup(self):
        """Create a subdispatcher."""
        await super().setup()

        log = self.cfg.get("log", None)
        root = self.root
        if root is None:
            raise RuntimeError("Not attached")

        if not log:
            self.sd = root.sub_at(self.cfg["path"])
            return

        from moat.lib.rpc import MsgSender  # noqa: PLC0415
        from moat.lib.rpc.loop import StreamLoop  # noqa: PLC0415

        a = StreamLoop(root, log + ">")
        b = StreamLoop(None, log + "<")
        a.attach_remote(b)
        b.attach_remote(a)
        await AC_use(self, a)
        xb = await AC_use(self, b)
        self.sd = MsgSender(xb)

    def handle(self, msg: Msg, rcmd: list[PathElem]) -> CoroutineType[Any, Any, None]:
        """Call via the subdispatcher."""
        return self.sd.handle(msg, rcmd)



"""
Serial port access apps.
"""

from __future__ import annotations

from typing import TYPE_CHECKING as _TC

if _TC:
    from .link import Link as Link
    from .msg import Msg as Msg
    from .raw import Raw as Raw

_imports = {
    "Link": "link",
    "Msg": "msg",
    "Raw": "raw",
}


def __getattr__(attr: str):
    try:
        mod = _imports[attr]
    except KeyError:
        raise AttributeError(attr) from None
    value = getattr(__import__(mod, globals(), None, (attr,), 1), attr)
    globals()[attr] = value
    return value


def __dir__():
    """Expose all lazy-loaded attributes for introspection."""
    return sorted(set(list(__all__) + [k for k in globals().keys() if not k.startswith("_")]))


__all__ = ["Link", "Msg", "Raw"]



"""
Error classes et al. for moat-lib-cmd.
"""

from __future__ import annotations

from moat.lib.micro import CancelledError
from moat.lib.proxy import as_proxy

from .const import (
    E_CANCEL,
    E_ERROR,
    E_MUST_STREAM,
    E_NO_CMD,
    E_NO_CMDS,
    E_NO_STREAM,
    E_SKIP,
    E_UNSPEC,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


@as_proxy("_NRdyErr")
class NotReadyError(RuntimeError):
    "An element of the command path was not ready"

    pass


@as_proxy("_SCmdErr")
class ShortCommandError(ValueError):
    "The command path was too short"

    pass


@as_proxy("_LCmdErr")
class LongCommandError(ValueError):
    "The command path was too long"

    pass


@as_proxy("_rErr")
class RemoteError(RuntimeError):
    "Some remote error that is not proxied"

    pass


class StreamError(RuntimeError):  # noqa: D101
    def __new__(cls, msg: Sequence[object] = ()):  # noqa: D102
        if len(msg) != 1:
            pass
        elif isinstance((m := msg[0]), int):
            if m >= 0:
                return Flow(m)
            elif m == E_UNSPEC:
                return super().__new__(StopMe)
            elif m == E_NO_STREAM:
                return super().__new__(NoStream)
            elif m == E_MUST_STREAM:
                return super().__new__(MustStream)
            elif m == E_SKIP:
                return super().__new__(SkippedData)
            elif m == E_NO_CMDS:
                return super().__new__(NoCmds)
            elif m == E_CANCEL:
                return CancelledError()
            elif m == E_ERROR:
                return super().__new__(RemoteError)
            elif m <= E_NO_CMD:
                return super().__new__(NoCmd, E_NO_CMD - m)
        elif isinstance(m, Exception):
            return m
        return super().__new__(cls, *msg)

    def __init__(self, msg=()):
        pass


class Flow(BaseException):
    "Flow control indication."

    def __init__(self, n):
        self.n = n


@as_proxy("_CSMErr")
class StopMe(StreamError):
    "Unspecified Stop"

    pass


@as_proxy("_CSDErr")
class SkippedData(StreamError):
    "Data skipped, took too long"

    pass


@as_proxy("_CNsErr")
class NoStream(StreamError):
    "No streaming support"

    pass


@as_proxy("_CNCsErr")
class NoCmds(StreamError):
    "No support for any commands"

    pass


@as_proxy("_CNCErr")
class NoCmd(StreamError):
    "Unknown command"

    pass


@as_proxy("_CWSErr")
class WantsStream(StreamError):
    "API: NoStream called on a streaming endpoint"

    pass


@as_proxy("_CMSErr")
class MustStream(StreamError):
    "Requires streaming support"

    pass


__all__ = [
    "Flow",
    "LongCommandError",
    "MustStream",
    "NoCmd",
    "NoCmds",
    "NoStream",
    "NotReadyError",
    "RemoteError",
    "ShortCommandError",
    "SkippedData",
    "StopMe",
    "StreamError",
    "WantsStream",
]



"""
Logging support for stream stacks.
"""

from __future__ import annotations

from moat.lib.micro import log
from moat.lib.rpc import B_FLAGSTR, wire2i_f

from .base import StackedBlk, StackedBuf, StackedMsg

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import Buffer, MutBuffer


class LogMsg(StackedMsg, StackedBuf, StackedBlk):
    """
    Log whatever messages cross this stack.

    This class implements all of StackedMsg/Buf/Blk.
    """

    # StackedMsg is first because MicroPython uses only the first class and
    # we get `cwr` and `crd` that way.

    def __init__(self, link, cfg):
        super().__init__(link, cfg)
        self.txt = cfg.get("txt", "S")
        self.decode = self.txt[0] == "!"
        if self.decode:
            self.txt = self.txt[1:]

    async def setup(self):  # noqa:D102
        log("X:%s start", self.txt)
        await super().setup()

    async def teardown(self):  # noqa:D102
        log("X:%s stop", self.txt)
        await super().teardown()

    def _repr(self, m, sub=None):
        if not isinstance(m, dict):
            return repr(m)
        res = []
        for k, v in m.items():
            if sub == k:
                res.append(f"{k}={self._repr(v)}")
            else:
                res.append(f"{k}={v!r}")
        return "{" + " ".join(res) + "}"

    def _repr_bang(self, m):
        m = m[:]
        i, fl = wire2i_f(m.pop(0))
        f = B_FLAGSTR[fl]
        if i >= 0:
            f += "+"
        f += str(i)
        return f, self._repr(m)

    async def send(self, m):  # noqa:D102
        if self.decode and isinstance(m, (list, tuple)) and m and isinstance(m[0], int):
            log("S:%s %s %s", self.txt, *self._repr_bang(m))
        else:
            log("S:%s %s%s", self.txt, "? " if self.decode else "", self._repr(m, "d"))
        try:
            res = await self.s.send(m)
        except BaseException as exc:
            log("S:%s stop %r", self.txt, exc)
            raise
        else:
            # log("S:%s =%s", self.txt, self._repr(res, "d"))
            return res

    async def recv(self):  # noqa:D102
        # log("R:%s", self.txt)
        try:
            msg = await self.s.recv()
        except BaseException as exc:
            log("R:%s stop %r", self.txt, exc)
            raise
        else:
            if self.decode and isinstance(msg, (list, tuple)) and msg and isinstance(msg[0], int):
                log("R:%s %s %s", self.txt, *self._repr_bang(msg))
            else:
                log("R:%s %s%s", self.txt, "? " if self.decode else "", self._repr(msg, "d"))
            return msg

    async def snd(self, m):  # noqa:D102
        log("SB:%s %r", self.txt, repr_b(m))
        try:
            return await self.s.snd(m)
        except BaseException as exc:
            log("SB:%s stop %r", self.txt, exc)
            raise

    async def rcv(self):  # noqa:D102
        # log("RB:%s", self.txt)
        try:
            msg = await self.s.rcv()
        except BaseException as exc:
            log("RB:%s stop %r", self.txt, exc)
            raise
        else:
            log("RB:%s %r", self.txt, repr_b(msg))
            return msg

    async def wr(self, data: Buffer) -> int:  # noqa:D102
        log("S:%s %r", self.txt, repr_b(data))
        try:
            res = await self.s.wr(data)
        except BaseException as exc:
            log("S:%s stop %r", self.txt, exc)
            raise
        else:
            # log("S:%s =%r", self.txt, res)
            return res

    async def rd(self, buf: MutBuffer) -> int:  # noqa:D102
        # log("R:%s %d", self.txt, len(buf))
        try:
            res = await self.s.rd(buf)
        except BaseException as exc:
            log("R:%s stop %r", self.txt, exc)
            raise
        else:
            log("R:%s %r", self.txt, repr_b(buf[:res]))
            return res

    async def cwr(self, buf: Buffer) -> None:  # noqa:D102
        log("SC:%s %r", self.txt, repr_b(buf))
        try:
            res = await self.s.cwr(buf)
        except BaseException as exc:
            log("SC:%s stop %r", self.txt, exc)
            raise
        else:
            # log("SC:%s =%r", self.txt, res)
            return res

    async def crd(self, buf: MutBuffer) -> int:  # noqa:D102
        # log("RC:%s %d", self.txt, len(buf))
        try:
            res = await self.s.crd(buf)
        except BaseException as exc:
            log("RC:%s stop %r", self.txt, exc)
            raise
        else:
            log("RC:%s %r", self.txt, repr_b(buf[:res]))
            return res


def repr_b(b):
    "show bytearray and memoryview as bytes"
    if isinstance(b, bytes):
        return b
    return bytes(b)


LogBuf = LogMsg
LogBlk = LogMsg



"""
This is a micropython-compatible MsgPack implementation.
It does not support
* a "default" fallback encoder
* encoding binary data to strings (which is legacy nonsense anyway)
* auto-encoding datetime to the timestamp extension
* the object_pairs hook

The encoder returns bytes.
The decoder yields messages.
You can also decode single messages.

The decoder returns binary data as memoryviews if they're larger
than the threshold (default -1: always copy). Extension objects always
get a memoryview and must decode or copy it.
"""

from __future__ import annotations

import struct
import sys
from io import BytesIO

from moat.util import OutOfData, attrdict
from moat.lib.micro import byte2utf8, const

from ._base import Codec as _Codec
from ._base import NoCodecError

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from ._base import ByteType, VarByteType

__all__ = ["Codec", "ExtType"]


class Codec(_Codec):
    "Extensible msgpack codec"

    def __init__(self, use_attrdict: bool = False, **kw):
        # TODO add keywords for msgpack enc/dec settings
        super().__init__(**kw)
        self.use_attrdict = use_attrdict
        self.__kw = kw

        self.stream = Unpacker(
            ext_hook=self._decode,  # pyright:ignore
            use_attrdict=use_attrdict,
        )

    def copy(self) -> Codec:
        "copy me"
        return Codec(use_attrdict=self.use_attrdict, **self.__kw)

    def encode(self, obj):
        "object > bytes"
        return packb(obj, default=self._encode)

    def _encode(self, obj):
        k, d = self.ext.encode(self, obj)
        return ExtType(k, d)

    def decode(self, data: ByteType):
        "bytes > object"
        return unpackb(
            data,
            ext_hook=self._decode,
            use_attrdict=self.use_attrdict,
        )

    def _decode(self, key, data):
        try:
            return self.ext.decode(self, key, data)
        except NoCodecError:
            return ExtType(key, data)

    def feed(self, data):
        "Add more bytes. Returns an iterator for the result."
        self.stream.feed(data)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.stream)

    def unfeed(self, buf: VarByteType | None) -> int:
        "Take from the decoder's buffer."
        return self.stream.unfeed(buf)


# cloned from https://github.com/msgpack/msgpack-python

# Parts of this have been modified to be compatble with micropython.


class UnpackException(Exception):
    "superclass, not raised"


class FormatError(UnpackException):
    "Error code read"


class ExtraData(ValueError):
    "too much data in buffer"

    def __init__(self, unpacked, extra):
        self.unpacked = unpacked
        self.extra = extra

    def __str__(self):
        return "unpack(b) received extra data."


class ExtType:
    """ExtType represents extension types in msgpack."""

    def __init__(self, code, data):
        if isinstance(data, memoryview):
            data = bytes(data)
        self.code = code
        self.data = data

    def __repr__(self):
        return f"Ext({self.code}:{self.data})"


_TYPE_IMMEDIATE = const(0)
_TYPE_ARRAY = const(1)
_TYPE_MAP = const(2)
_TYPE_RAW = const(3)
_TYPE_BIN = const(4)
_TYPE_EXT = const(5)

DEFAULT_RECURSE_LIMIT = 20


_MSGPACK_HEADERS: dict[int, tuple[int, str] | tuple[int, str, int]] = {
    0xC4: (1, "", _TYPE_BIN),
    0xC5: (2, ">H", _TYPE_BIN),
    0xC6: (4, ">I", _TYPE_BIN),
    0xC7: (2, "Bb", _TYPE_EXT),
    0xC8: (3, ">Hb", _TYPE_EXT),
    0xC9: (5, ">Ib", _TYPE_EXT),
    0xCA: (4, ">f"),
    0xCB: (8, ">d"),
    0xCC: (1, ""),
    0xCD: (2, ">H"),
    0xCE: (4, ">I"),
    0xCF: (8, ">Q"),
    0xD0: (1, "b"),
    0xD1: (2, ">h"),
    0xD2: (4, ">i"),
    0xD3: (8, ">q"),
    0xD4: (1, "b1s", _TYPE_EXT),
    0xD5: (2, "b2s", _TYPE_EXT),
    0xD6: (4, "b4s", _TYPE_EXT),
    0xD7: (8, "b8s", _TYPE_EXT),
    0xD8: (16, "b16s", _TYPE_EXT),
    0xD9: (1, "", _TYPE_RAW),
    0xDA: (2, ">H", _TYPE_RAW),
    0xDB: (4, ">I", _TYPE_RAW),
    0xDC: (2, ">H", _TYPE_ARRAY),
    0xDD: (4, ">I", _TYPE_ARRAY),
    0xDE: (2, ">H", _TYPE_MAP),
    0xDF: (4, ">I", _TYPE_MAP),
}


class Unpacker:
    """
    Manager for buffered and streamed unpacking.
    """

    def __init__(
        self,
        ext_hook=ExtType,
        use_attrdict=False,
        min_memview_len=-1,
    ):
        #: array of bytes fed.
        self._buffer = bytearray()
        #: Which position we currently reads
        self._buf_pos = 0

        self._ext_hook = ext_hook
        self._min_memview_len = min_memview_len
        self._use_attrdict = use_attrdict

    def feed(self, data):
        "set the buffer"
        if self._buffer and self._buf_pos < len(self._buffer):
            if self._buf_pos == 0:
                self._buffer += data
                return
            data = self._buffer[self._buf_pos :] + data
        elif isinstance(data, memoryview):
            data = bytearray(data)
        self._buffer = data
        self._buf_pos = 0

    def unfeed(self, buf: VarByteType | None) -> int:
        "take from the buffer"
        if not self._buffer:
            return 0
        i = self._buf_pos
        n = min(len(buf) if buf is not None else 999, len(self._buffer) - i)
        if n == 0:
            return 0
        i_n = i + n
        if buf is not None:
            buf[:n] = self._buffer[i:i_n]
        self._buf_pos = i_n
        return n

    def has_extradata(self):
        "are there extra data in the buffer?"
        return self._buf_pos < len(self._buffer)

    def get_extradata(self):
        "return extra data, if any"
        return self._buffer[self._buf_pos :]

    def _read(self, n):
        # (int) -> bytearray
        i = self._buf_pos
        i_n = i + n
        if i_n > len(self._buffer):
            raise OutOfData

        ret = self._buffer[i:i_n]
        self._buf_pos = i + len(ret)
        return ret

    def _reserve(self, n):
        remain_bytes = len(self._buffer) - self._buf_pos - n

        # Fast path: buffer has n bytes already
        if remain_bytes >= 0:
            return

        raise OutOfData

    def _read_header(self):
        typ = _TYPE_IMMEDIATE
        n = 0
        obj = None
        self._reserve(1)
        b = self._buffer[self._buf_pos]
        self._buf_pos += 1
        if b & 0b10000000 == 0:  # x00-x7F
            obj = b
        elif b & 0b11100000 == 0b11100000:  # xE0-xFF
            obj = -1 - (b ^ 0xFF)
        elif b & 0b11100000 == 0b10100000:  # xA0-xBF
            n = b & 0b00011111
            typ = _TYPE_RAW
            obj = self._read(n)
        elif b & 0b11110000 == 0b10010000:  # x90-x9F
            n = b & 0b00001111
            typ = _TYPE_ARRAY
        elif b & 0b11110000 == 0b10000000:  # x80-x8F
            n = b & 0b00001111
            typ = _TYPE_MAP
        elif b == 0xC0:
            obj = None
        elif b == 0xC1:
            raise FormatError("unused code")
        elif b == 0xC2:
            obj = False
        elif b == 0xC3:
            obj = True
        elif b <= 0xC6:
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size)
            if len(fmt) > 0:
                n = struct.unpack_from(fmt, self._buffer, self._buf_pos)[0]
            else:
                n = self._buffer[self._buf_pos]
            self._buf_pos += size
            obj = self._read(n)
        elif b <= 0xC9:
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size)
            L, n = struct.unpack_from(fmt, self._buffer, self._buf_pos)
            self._buf_pos += size
            obj = self._read(L)
        elif b <= 0xD3:
            size, fmt = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 2-tuple for these keys
            self._reserve(size)
            if len(fmt) > 0:
                obj = struct.unpack_from(fmt, self._buffer, self._buf_pos)[0]
            else:
                obj = self._buffer[self._buf_pos]
            self._buf_pos += size
        elif b <= 0xD8:
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size + 1)
            n, obj = struct.unpack_from(fmt, self._buffer, self._buf_pos)
            self._buf_pos += size + 1
        elif b <= 0xDB:
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size)
            if len(fmt) > 0:
                (n,) = struct.unpack_from(fmt, self._buffer, self._buf_pos)
            else:
                n = self._buffer[self._buf_pos]
            self._buf_pos += size
            obj = self._read(n)
        elif b <= 0xDD:
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size)
            (n,) = struct.unpack_from(fmt, self._buffer, self._buf_pos)
            self._buf_pos += size
        else:  # if b <= 0xDF:  # can't be anything else
            size, fmt, typ = _MSGPACK_HEADERS[b]  # type: ignore[misc]  # always 3-tuple for these keys
            self._reserve(size)
            (n,) = struct.unpack_from(fmt, self._buffer, self._buf_pos)
            self._buf_pos += size
        return typ, n, obj

    def unpack(self):
        "extract one (top-level) item from the buffer"
        i = self._buf_pos
        try:
            res = self._unpack()
        except BaseException:
            self._buf_pos = i
            raise

        # Buffer management: chop off the part we've read
        self._buffer = self._buffer[self._buf_pos :]
        self._buf_pos = 0
        return res

    def _unpack(self):
        typ, n, obj = self._read_header()

        if typ == _TYPE_ARRAY:
            ret = []
            for _ in range(n):
                ret.append(self._unpack())
            # if self._list_hook is not None:
            # ret = self._list_hook(ret)
            return ret
        if typ == _TYPE_MAP:
            ret = attrdict() if self._use_attrdict else dict()
            for _ in range(n):
                key = self._unpack()
                if type(key) is str and hasattr(sys, "intern"):
                    key = sys.intern(key)
                elif isinstance(key, (bytearray, memoryview)):
                    key = bytes(key)
                ret[key] = self._unpack()
            # if self._object_hook is not None:
            # ret = self._object_hook(ret)
            return ret
        if typ == _TYPE_RAW:
            return byte2utf8(cast(bytes, obj))
        if typ == _TYPE_BIN:
            if self._min_memview_len < 0 and len(cast(bytes, obj)) < self._min_memview_len:
                obj = bytes(cast(bytes, obj))
            return obj
        if typ == _TYPE_EXT:
            return self._ext_hook(n, cast(bytes, obj))
        # assert typ == _TYPE_IMMEDIATE
        return obj

    def __aiter__(self):
        return self

    def __anext__(self):
        return self.unpack()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.unpack()
        except OutOfData:
            raise StopIteration  # noqa:B904,RUF100


class Packer:
    """
    Manager for buffered and streamed packing.
    """

    def __init__(
        self,
        # unicode_errors=None,
        default=None,
    ):
        self._buffer = BytesIO()
        # self._unicode_errors = unicode_errors or "strict"
        self._default = default

    def _pack(self, obj, default=None):
        # Warning, does not deal with recursive data structures
        # (except by running out of memory)
        list_types = (list, tuple)
        if default is None:
            default = self._default

        todo = [obj]

        # shorter bytecode
        def wp(*x):
            return self._buffer.write(struct.pack(*x))

        wb = self._buffer.write
        is_ = isinstance

        _ndefault = default
        while todo:
            _default, _ndefault = _ndefault, default
            obj = todo.pop()
            if obj is None:
                wb(b"\xc0")
                continue
            if isinstance(obj, bool):
                wb(b"\xc3" if obj else b"\xc2")
                continue
            if isinstance(obj, int):
                if obj >= 0:
                    if obj < 0x80:
                        wp("B", obj)
                        continue
                    if obj <= 0xFF:
                        wp("BB", 0xCC, obj)
                        continue
                    if obj <= 0xFFFF:
                        wp(">BH", 0xCD, obj)
                        continue
                    if obj <= 0xFFFFFFFF:
                        wp(">BI", 0xCE, obj)
                        continue
                    if obj <= 0xFFFFFFFFFFFFFFFF:
                        wp(">BQ", 0xCF, obj)
                        continue
                else:
                    if obj >= -0x20:
                        wp("b", obj)
                        continue
                    if obj >= -0x80:
                        wp(">Bb", 0xD0, obj)
                        continue
                    if obj >= -0x8000:
                        wp(">Bh", 0xD1, obj)
                        continue
                    if obj >= -0x80000000:
                        wp(">Bi", 0xD2, obj)
                        continue
                    if obj >= -0x8000000000000000:
                        wp(">Bq", 0xD3, obj)
                        continue
                if _default:
                    res = _default(obj)
                    if res is not None:
                        todo.append(res)
                        _ndefault = False
                        continue
                raise OverflowError("Integer value out of range")
            if is_(obj, (bytes, bytearray, memoryview)):
                # XXX we have a problem if memoryview.itemsize != 1
                n = len(obj)
                self._pack_bin_header(n)
                wb(obj)
                continue
            if is_(obj, str):
                obj = obj.encode("utf-8")  # , self._unicode_errors)
                n = len(obj)
                if n <= 0x1F:
                    wb(struct.pack("B", 0xA0 + n))
                elif n <= 0xFF:
                    wb(struct.pack(">BB", 0xD9, n))
                elif n <= 0xFFFF:
                    wb(struct.pack(">BH", 0xDA, n))
                else:
                    wb(struct.pack(">BI", 0xDB, n))
                wb(obj)
                continue
            if is_(obj, float):
                # if self._use_float:
                wp(">Bf", 0xCA, obj)
                # else:
                # wp(">Bd", 0xCB, obj)
                continue
            if is_(obj, ExtType):
                code = obj.code
                data = obj.data
                L = len(data)
                if L == 1:
                    wb(b"\xd4")
                elif L == 2:
                    wb(b"\xd5")
                elif L == 4:
                    wb(b"\xd6")
                elif L == 8:
                    wb(b"\xd7")
                elif L == 16:
                    wb(b"\xd8")
                elif L <= 0xFF:
                    wp(">BB", 0xC7, L)
                elif L <= 0xFFFF:
                    wp(">BH", 0xC8, L)
                else:
                    wp(">BI", 0xC9, L)
                wp("b", code)
                wb(data)
                continue
            if is_(obj, list_types):
                n = len(obj)
                if n <= 0x0F:
                    wb(struct.pack("B", 0x90 + n))
                elif n <= 0xFFFF:
                    wb(struct.pack(">BH", 0xDC, n))
                else:
                    wb(struct.pack(">BI", 0xDD, n))
                while n > 0:
                    n -= 1
                    todo.append(obj[n])
                continue
            if is_(obj, dict):
                n = len(obj)
                if n <= 0x0F:
                    wb(struct.pack("B", 0x80 + n))
                elif n <= 0xFFFF:
                    wb(struct.pack(">BH", 0xDE, n))
                else:
                    wb(struct.pack(">BI", 0xDF, n))
                for k, v in obj.items():
                    todo.append(v)
                    todo.append(k)
                continue
            if _default:
                res = _default(obj)
                if res is not None:
                    _ndefault = False
                    todo.append(res)
                    continue
            raise TypeError(f"Cannot serialize {obj!r}")

    def pack(self, obj):
        "Packs a single data item. Returns the bytes."
        try:
            self._pack(obj)
            return self._buffer.getvalue()
        finally:
            self._buffer = BytesIO()  # always reset

    def _pack_bin_header(self, n):
        wb = self._buffer.write
        if n <= 0xFF:
            return wb(struct.pack(">BB", 0xC4, n))
        if n <= 0xFFFF:
            return wb(struct.pack(">BH", 0xC5, n))
        return wb(struct.pack(">BI", 0xC6, n))


def packb(o, **kwargs):
    """
    Pack object `o` and return packed bytes

    See :class:`Packer` for options.
    """
    return Packer(**kwargs).pack(o)


def unpackb(packed, **kwargs):
    """
    Unpack an object from `packed`.

    Raises ``ExtraData`` when *packed* contains extra bytes.
    Raises ``OutOfData`` when *packed* is incomplete.
    Raises ``FormatError`` when *packed* is not valid msgpack.
    Other exceptions can be raised during unpacking.

    See :class:`Unpacker` for options.
    """
    unpacker = Unpacker(**kwargs)
    unpacker.feed(packed)
    try:
        res = unpacker.unpack()
        if unpacker.has_extradata():
            raise ExtraData
        return res
    except OutOfData:
        raise OutOfData("incomplete") from None
    raise RuntimeError("No way")



"""
Base classes for command handlers.
"""

from __future__ import annotations

import sys
from contextlib import contextmanager

from moat.util import NotGiven
from moat import DOC
from moat.lib.micro import ACM, AC_exit, L, TaskGroup
from moat.lib.path import Path
from moat.lib.stream import Base
from moat.util.exc import ungroup

from .const import SD_BOTH, SD_IN, SD_NONE, SD_OUT
from .errors import NotReadyError, ShortCommandError

from collections.abc import Mapping, MutableMapping
from typing import TYPE_CHECKING, cast, overload

if TYPE_CHECKING or DOC:
    from types import (
        CoroutineType,  # noqa:TC003
        EllipsisType,  # noqa:TC003
    )

    from moat.lib.path import PathElem

    from .msg import Msg  # noqa:TC001

    from collections.abc import Awaitable, Callable, Iterator
    from typing import Any, Literal, Protocol, Self

    Key = str | int | bool
    OptDict = MutableMapping[str, Any] | None
    _SdrPath = tuple["MsgRoot", Path]

    class MsgRoot(Protocol):
        """Interface required by :class:`MsgSender`."""

        def handle(self, msg: Msg, rcmd: list[PathElem]) -> CoroutineType[Any, Any, None]: ...

        def find_handler(self, path: Path, cmd: bool = False) -> _SdrPath | Callable: ...

else:
    MsgRoot = object
    _SdrPath = object

__all__ = ["BaseMsgHandler", "MsgHandler", "MsgSender"]


class Caller:
    """
    This is the Wrapper returned by `MsgSender.cmd`. It can be *await*ed
    (direct RPC call), or used as an async context manager (data streaming).

    The context manager yields an async iterator for incoming messages.

    .. seealso: moat.lib.rpc.Msg

    You should not instantiate this class directly.
    """

    _qlen = 0

    def __init__(
        self,
        sender: BaseMsgHandler | MsgRoot,
        data: tuple[str | Path, list[Any] | tuple[Any, ...], dict[str, Any]],
        _list: bool | EllipsisType | None = NotGiven,
    ):
        """
        Args:
            sender: the MsgSender on which we call :py:meth:`MsgSender.handle` to get the result.
            data: (cmd,args,kw) tuple.
            _list: can be
                - NotGiven: return the message object
                - True: return a list
                - False: return a dict
                - None: best effort (single/map/list/message object)
        """
        self.data = data
        self.sender = sender
        self._dir = SD_NONE
        self._list = _list

    def __await__(self):
        "makes this object awaitable, CPython"
        return self._call().__await__()

    def __iter__(self):
        "makes this object awaitable, MicroPython"
        # this depends on µPy doing the right thing
        return self._call()

    async def _call(self):
        "helper for __await__ that calls the remote side"
        from .msg import Msg  # noqa: PLC0415

        cmd, args, kw = self.data
        if isinstance(cmd, str):
            cmd = Path.build((cmd,))
        msg = Msg.Call(cmd, list(args), kw)

        await self.sender.handle(msg, msg.rcmd)
        await msg.wait_replied()

        if self._list is NotGiven:
            return msg

        if self._list is True:
            # always return a list
            if msg.kw:
                raise ValueError("has dict", msg)
            return msg.args

        if self._list is False:
            # always return a dict
            if msg.args:
                raise ValueError("has args", msg)
            return msg.kw

        if kw := msg.kw:
            if msg.args:
                # return the message object if both kw and args are set
                return msg
            # return a dict if only kw is set
            return kw

        args = msg.args
        if len(args) == 1:
            # return a single arg directly
            return args[0]
        # otherwise return all of them (if any)
        return args

    async def __aenter__(self) -> Msg:
        acm = ACM(self)
        try:
            if not self._dir:
                self._dir = SD_BOTH

            from .msg import Msg  # noqa: PLC0415

            cmd, args, kw = self.data
            if isinstance(cmd, str):
                cmd = Path.build((cmd,))
            m1 = Msg.Call(cmd, list(args), kw)

            tg = await acm(TaskGroup())
            m2 = await acm(m1.ensure_remote())
            # m2 is the one with the command data
            tg.start_soon(self.sender.handle, m2, m2.rcmd)
            await acm(m1._stream_call(self._dir))  # noqa: SLF001
            return m1
        except BaseException as exc:
            try:
                await AC_exit(self, type(exc), exc, None)
            except BaseException as ex:
                ex = ungroup.one(ex)
                if ex is not ungroup.one(exc):
                    raise
            raise

    async def __aexit__(self, *err):
        return await AC_exit(self, *err)

    def stream(self, size: int = 42) -> Self:
        """mark as streaming bidirectionally (the default)

        Args:
            size: length of the incoming queue
        """
        assert not self._dir
        self._dir = SD_BOTH
        self._qlen = size
        return self

    def stream_in(self, size: int = 42) -> Self:
        """mark as only streaming in.

        Args:
            size: length of the incoming queue
        """
        assert not self._dir
        self._dir = SD_IN
        self._qlen = size
        return self

    def stream_out(self) -> Self:
        """mark as only streaming out"""
        assert not self._dir
        self._dir = SD_OUT
        return self


class BaseMsgHandler:
    """
    Somewhat-abstract superclass for anything that accepts messages.
    """

    def handle(self, msg: Msg, rcmd: list[PathElem]) -> CoroutineType[Any, Any, None]:
        """
        Handle this message stream.

        Args:
            msg: the message to process.
            rcmd: the inverted command prefix. Hierarchical handlers chop an
                  element off the end.
        """
        raise NotImplementedError


class MsgSender(BaseMsgHandler):
    """
    This class is the client-side API of the MoaT Command multiplexer.

    It is typically returned by the :meth:`MsgHandler.sender` method. The
    MsgHandler's :meth:`handle` method processes messages from the remote side,
    while this class accepts local messages and forwards them to the
    remote.
    """

    Caller_: type[Caller] = Caller

    def __init__(self, root: MsgRoot):
        """
        This class accepts client-side MoaT-cmd calls and turns them into
        messages.
        """
        self._root = root

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        pass

    @property
    def sender(self) -> MsgSender:
        """Return a sender adapter for compatibility with handler roots."""
        return self

    @property
    def root(self) -> MsgRoot:  # noqa: D102
        return self._root

    def set_root(self, root: MsgRoot):  # noqa: D102
        if type(self) is not MsgSender:
            raise RuntimeError("not in a subclass")
        assert not isinstance(root, MsgSender)
        self._root = root

    def handle(self, msg: Msg, rcmd: list[PathElem]) -> CoroutineType[Any, Any, None]:
        """
        Redirect to the underlying command handler.
        """
        return self.root.handle(msg, rcmd)

    def cmd(self, cmd: Path, *a: Any, **kw: Any) -> Caller:
        """
        Run the command at this path.

        The result of this call can be used both as an awaitable (straight
        method call) and a context manager (streaming):

        >>> res = await hdl.cmd("foo",42)
        >>> async with hdl.cmd("bar",x=21) as m:
        ...     async for msg, in m:
        ...         m.send(msg*2)


        """
        return self.Caller_(self, (cmd, a, kw))

    @overload
    def sub_at(self, prefix: Path, caller=None, cmd: Literal[False] = False) -> MsgSender: ...

    @overload
    def sub_at(self, prefix: Path, caller=None, *, cmd: Literal[True]) -> MsgSender | Callable: ...

    @overload
    def sub_at(self, prefix: Path, caller=None, *, cmd: bool) -> MsgSender: ...

    def sub_at(self, prefix: Path, caller=None, cmd: bool = False) -> MsgSender | Callable:
        """
        Resolve the given prefix, as far as possible.

        Returns a Callable, typically a :class:`SubMsgSender` — but
        possibly a method if the path can be resolved and *cmd* is `True`.
        """
        res = self.root.find_handler(prefix, cmd=cmd)
        if isinstance(res, tuple):
            root: MsgRoot
            rem: Path
            root, rem = cast(_SdrPath, res)
            if rem:
                return SubMsgSender(root, rem, caller=caller or self.Caller_)
            return MsgSender(root)
        return res

    def cfg_at(self, p: Path):
        "returns a CfgStore object that accesses the config at this subpath"

        from moat.lib.rpc import CfgStore  # noqa: PLC0415

        return CfgStore(self, p)

    def add_sub(self, elem: str):
        """
        Ensures that ``self.ELEM`` is a SubMsgSender.
        """
        if hasattr(self, elem):
            sb = getattr(self, elem)
            assert isinstance(sb, SubMsgSender)
        else:
            sb = self.sub_at(Path.build((elem,)))
            setattr(self, elem, sb)
        return sb

    def __getattr__(self, x: str) -> MsgSender:
        """
        Return a sub-sender for this path element.
        """
        return self.sub_at(Path.build((x,)))

    __getitem__ = __getattr__


class SubMsgSender(MsgSender):
    """
    This `MsgSender` subclass auto-prefixes a path to all calls.

    Thus,

    >>> async with some_handler() as h:
    ...     h.add_sub("a")
    ...     await h.a.b.c()

    is the same as

    >>> async with some_handler() as h, h.sub_at(P("a.b")) as hab:
    ...     await hab.c()

    The additional context also verifies that the remote link is up
    (or rather, waits for it to *be* up).

    An empty path is OK.
    """

    def __init__(self, root: MsgRoot, path: Path, caller=None):
        """
        Setup.
        """
        super().__init__(root)
        self._path = path
        self._rpath = list(path)
        self._rpath.reverse()
        if caller is not None:
            self.Caller_ = caller

    def __repr__(self):
        return f"<{self.__class__.__name__}:{self.root} {self._path}"

    @property
    def path(self):
        "The subpath accessible with this object"
        return self._path

    async def __aenter__(self):
        "Ensure that the called object is ready for service"
        try:
            msg = await self.cmd(Path.build(("rdy_",)))
        except KeyError:
            pass  # can't do it. Oh well.
        else:
            if msg[0]:
                raise NotReadyError(self)
        return await super().__aenter__()

    def handle(self, msg: Msg, rcmd: list[PathElem]) -> CoroutineType[Any, Any, None]:
        """
        Forward the message, as directed by :py:attr:`path`.
        """
        rcmd.extend(self._rpath)
        return self._root.handle(msg, rcmd)

    def cmd(self, cmd: Path, *a: Any, **kw: Any) -> Caller:
        """
        Run a command.

        The result of this call can be used both as an Awaitable (straight
        method call) and an async context manager (streaming):

        >>> res = await hdl.cmd(P("foo"),42)
        >>> async with hdl.cmd(P("times_two")) as m:
        ...     async for msg, in m:
        ...         m.send(msg*2)
        """
        # The path is modified later, when the Caller object
        # calls our .handle method
        return Caller(self, (cmd, a, kw))

    def stream(self, *a, **kw):
        return self.cmd(Path(), *a, **kw).stream()

    def stream_in(self, *a, **kw):
        return self.cmd(Path(), *a, **kw).stream_in()

    def stream_out(self, *a, **kw):
        return self.cmd(Path(), *a, **kw).stream_out()

    def __call__(self, *a: Any, _list: bool | EllipsisType | None = None, **kw: Any) -> Caller:
        """
        Process a direct call.

        This makes some kind of best effort to unpack the result.
        If ``_list`` is False, always returns a dict.
        If ``_list`` is True, always returns a list.

        Args:
            _list: Unpacking mode.
        """
        return self.Caller_(self.root, (self._path, a, kw), _list=_list)

    @overload
    def sub_at(self, prefix: Path, caller=None, cmd: Literal[False] = False) -> SubMsgSender: ...

    @overload
    def sub_at(self, prefix: Path, caller=None, *, cmd: Literal[True]) -> SubMsgSender: ...

    @overload
    def sub_at(self, prefix: Path, caller=None, *, cmd: bool) -> SubMsgSender: ...

    def sub_at(self, prefix: Path, caller=None, cmd: bool = False) -> SubMsgSender:
        """
        Returns a SubMsgSender
        """
        cmd  # noqa:B018
        return SubMsgSender(self.root, self._path + prefix, caller=caller or self.Caller_)

    def __getattr__(self, x):
        """
        Returns a SubMsgSender for this name.
        """
        return self.sub_at(Path.build((x,)))

    __getitem__ = __getattr__


class MsgHandler(Base, BaseMsgHandler, MsgRoot):
    """
    Message dispatching, using a Path-based prefix.

    Implement ``cmd(self, *a, *kw)`` and/or ``cmd_NAME(self, *a, *kw)``
    for simple method calls.

    Implement ``stream(self, msg)`` and/or ``stream_NAME(self,
    msg)`` for streamed calls.

    Use ``doc`` or ``doc_NAME`` for basic call introspection.

    Set the SKIP_RDY class attribute to allow readiness checks for
    subcommands to pass through even if the handler itself is not ready.
    """

    SKIP_RDY = False

    @property
    def root(self):  # noqa: D102
        return self

    @property
    def sender(self):
        """
        If you implement a ``MsgHandler`` that controls a remote link,
        override this method to return a MsgSender that transmits to the
        remote side.

        Raises `NotImplementedError`.
        """
        if sys.implementation.name == "micropython":
            return None
        raise NotImplementedError

    @contextmanager
    def delegate(self, path: Path, service: MsgHandler) -> Iterator[Self]:
        """
        Tell this handler to delegate messages that start with this path
        to that handler.

        Args:
            path: the prefix to delegate
            service: the service to call

        This is a context manager.
        """
        if len(path) == 0:
            raise RuntimeError(f"{path}: cannot be empty")
        name = f"sub_{path[0]}"
        if len(path) > 1:
            sub = getattr(self, name)  # must exist
            with sub.delegate(path[1:], service):
                yield self
        else:
            if hasattr(self, name):
                raise RuntimeError(f"{name}: already known")
            setattr(self, name, service)
            try:
                yield self
            finally:
                delattr(self, name)

    async def handle(self, msg: Msg, rcmd: list[PathElem]):
        """
        Process the message.

        * If the command is ``XX.doc_``, return ``self.doc_XX``. This should be
          a class attribute.

        * If the command ends with ``rdy_``, check for readiness before
          further processing (if any).

        * Multi-step commands ``XX.*`` are handled by the ``sub_XX`` method.
          (If that is a class instance: call its ``handle`` method.)

        * Non-streaming commands are processed by ``cmd_XX`` if it exists.

        * Otherwise, commands are processed by ``stream_XX``.

        * If that doesn't exist, raise a `KeyError`.

        """
        # Process direct calls.
        if not rcmd:
            if not msg.can_stream and (cmd := getattr(self, "cmd", None)) is not None:
                return await msg.call_simple(cmd)
            elif (cmd := getattr(self, "stream", None)) is not None:
                return await msg.call_stream(cmd)
            else:
                raise ShortCommandError(msg.cmd)

        # Process requests for documentation.
        if len(rcmd) <= 2 and rcmd[0] == "doc_":
            if msg.args or msg.kw:
                raise TypeError("doc", msg.args, msg.kw)
            if (
                doc := getattr(self, f"doc_{rcmd[1]}" if len(rcmd) > 1 else "doc", None)
            ) is not None:
                return await msg.result(doc)

        # Process command handlers of this class.
        if len(rcmd) == 1:
            if not msg.can_stream and (cmd := getattr(self, f"cmd_{rcmd[0]}", None)) is not None:
                return await msg.call_simple(cmd)
            if (cmd := getattr(self, f"stream_{rcmd[0]}", None)) is not None:
                return await msg.call_stream(cmd)

        # Neither of the above.
        # First test if it's a readiness check.
        is_rdy = False
        if rcmd[0] == "rdy_":
            if (
                L
                and not getattr(self, "SKIP_RDY", False)
                and hasattr(self, "wait_ready")
                and await self.wait_ready(wait=True)
            ):
                raise NotReadyError(msg.cmd, rcmd)
            is_rdy = True

        # Find a subcommand.
        scmd = rcmd.pop()
        if (sub := self.find_sub(scmd)) is not None:
            sub = getattr(sub, "handle", sub)
            sub = cast("Callable[[Msg, list[PathElem]], Awaitable[None]]", sub)
            return await sub(msg, rcmd)

        if is_rdy:
            return await msg.result(None)

        raise KeyError(
            scmd,
            self.cfg_name,
            msg.cmd,
            list(subs.keys()) if isinstance((subs := getattr(self, "sub", None)), Mapping) else (),
        )

    def find_sub(self, scmd: PathElem) -> BaseMsgHandler | Callable | None:
        """
        Resolve a subcommand.

        Args:
            scmd: the subcommand to look up.

        The default implementation uses ``sub_{scmd}`` attributes.
        """
        if not isinstance(scmd, str):
            return None
        return cast("BaseMsgHandler | Callable | None", getattr(self, f"sub_{scmd}", None))

    def find_handler(self, path, cmd: bool = False) -> tuple[MsgHandler, Path] | Callable:
        """
        Do a path lookup and find a suitable subcommand.

        This is a shortcut finder that returns either a subroot+prefix
        tuple or something that's callable directly. Implementing the
        latter is TODO.
        """
        cmd  # noqa:B018
        return self, path

    doc_dir_ = dict(
        _d="directory",
        v="bool:verbose",
        _r=dict(c=["str:commands"], d=["str:modules"], j="bool:callable"),
    )

    async def cmd_dir_(self, v=True):
        """
        Rudimentary introspection. Returns a dict with
        a list of available commands @c,
        iterators @i, and submodules @d.
        j=True if callable directly.

        If @v ("visible") is set (the default),
        this does not return hidden commands.
        """
        c = []
        s = []
        res = {}

        for k in dir(self):
            if v is (k[-1] == "_"):
                continue
            if k.startswith("cmd_"):
                c.append(k[4:])
            elif k.startswith("stream_"):
                s.append(k[7:])
            elif k == "cmd":
                res["C"] = True
            elif k == "stream":
                res["S"] = True
        if c:
            res["c"] = c
        if s:
            res["s"] = s
        return res

    async def check_rdy(self, msg: Msg, rcmd: list[PathElem]):
        """
        Basic boilerplate command for handling basic readiness checks in
        `handle` and similar methods.
        """
        if not L or not rcmd or rcmd[0] != "rdy_":
            return
        if await self.wait_ready(wait=msg.get("wait", True)):
            raise NotReadyError(msg.cmd, rcmd)

    if L:
        doc_rdy_ = dict(_d="check readiness", w="bool:wait for it?")

        def cmd_rdy_(self, w=True) -> Awaitable:
            """
            Check if / wait for readiness.

            See :meth:`wait_ready` for return values.
            """
            return self.wait_ready(wait=w)

        async def wait_ready(self, wait: bool = True) -> bool | None:
            """
            Check if this handler is ready.

            Args:
                wait: whether to sleep until ready.

            Returns:
                * ``True`` if stopped
                * ``False`` if it already running
                * ``None`` if the command has (or would have, if
                  *wait* is False) waited for it to become ready.
            """
            wait  # noqa:B018
            return False



"""
Access a satellite's SPI bus.
"""

from __future__ import annotations

try:
    import machine
except ImportError:
    from moat.micro._test import machine

from moat.lib.micro import Lock, to_thread
from moat.lib.rpc import BaseCmd

# Per-bus cache for hardware SPI: bus_id -> [bus, lock, owner]
_bus_cache: dict[int, list] = {}


class Cmd(BaseCmd):
    """
    This command implements basic access to a SPI bus.

    Parameters:

        id(int|None): None  # uses soft SPI if None
        sck(int): clock pin
        mosi(int): output / MOSI (master out, slave in)
        miso(int): input / MISO (master in, slave out)
        f(int): frequency(Hz)
        pol(bool): clock polarity (0 or 1)
        pha(bool): clock phase (0 or 1)
        cs(int): chip select pin
        cs_low(bool): chip select is active low? (default: `True`)

    Device operations require a CS (chip select) pin path.
    The driver controls CS: active low before operation, high after.
    """

    _bus = None

    doc = dict(
        _c=dict(_d="SPI driver"),
        id="int|None:hardware bus ID",
        sck="int:clock pin",
        mosi="int:MOSI pin",
        miso="int:MISO pin",
        f="int:frequency",
        cs="int:CS pin",
        cs_low="bool:CS pin is active low",
        pol="int:polarity(0)",
        pha="int:phase(0)",
    )

    async def setup(self):
        """
        Open a bus.
        """
        await super().setup()
        self._setup()

    async def reload(self):
        "reconfigured"
        self._teardown()
        self._setup()
        await super().reload()

    def _setup(self):
        cfg = self.cfg

        # CS pin handling
        if "cs" in cfg:
            self.cs_pin = machine.Pin(cfg["cs"], machine.Pin.OUT)
            self.cs_low = cfg.get("cs_low", True)
            # Initialize CS to inactive state
            self.cs_pin.value(not self.cs_low)
        else:
            self.cs_pin = None
            self.cs_low = True

        # Store SPI config for reinit
        self._spi_cfg = {
            "baudrate": cfg.get("f", 1000000),
            "polarity": cfg.get("pol", 0),
            "phase": cfg.get("pha", 0),
        }

        sck = machine.Pin(cfg["sck"])
        mosi = machine.Pin(cfg["mosi"])
        miso = machine.Pin(cfg["miso"])

        bus_id = cfg.get("id", None)
        self._bus_id = bus_id

        if bus_id is None:
            # Software SPI - each instance is independent
            self._bus = machine.SoftSPI(sck=sck, mosi=mosi, miso=miso, **self._spi_cfg)
            self.lock = Lock()
            self._cache = None
        else:
            # Hardware SPI - shared bus with reinit on owner change
            if bus_id not in _bus_cache:
                bus = machine.SPI(bus_id, sck=sck, mosi=mosi, miso=miso, **self._spi_cfg)
                _bus_cache[bus_id] = [bus, Lock(), None]
            self._cache = _bus_cache[bus_id]
            self._bus = self._cache[0]
            self.lock = self._cache[1]

    async def teardown(self):
        "shutdown"
        self._teardown()
        await super().teardown()

    def _teardown(self):
        if self._cache is not None:
            # Hardware SPI - clear ownership on shared bus
            if self._cache[2] is self:
                self._cache[2] = None
                self._bus.deinit()
            self._bus = None
        else:
            # Software SPI - we own it, always deinit
            self._bus.deinit()
            self._bus = None

    def _ensure_config(self):
        """Reinit bus if another driver used it since we did."""
        if self._cache is not None and self._cache[2] is not self:
            self._bus.init(**self._spi_cfg)
            self._cache[2] = self

    async def _with_cs(self, func):
        """
        Execute func with CS asserted.
        """
        async with self.lock:
            self._ensure_config()
            if self.cs_pin is not None:
                self.cs_pin.value(self.cs_low)  # Assert CS (active state)
            try:
                return await func()
            finally:
                if self.cs_pin is not None:
                    self.cs_pin.value(not self.cs_low)  # Deassert CS

    doc_r = dict(
        _d="read",
        n="int:nbytes",
        wr="int:write byte(0)",
    )

    async def cmd_r(self, n: int, wr: int = 0) -> bytes:
        """
        Read @n bytes from SPI while writing @wr byte.

        Args:
            n: number of bytes to read
            wr: no-op byte to write while reading (default 0x00)
        """

        async def _run():
            return await to_thread(self._bus.read, n, wr)

        return await self._with_cs(_run)

    doc_w = dict(
        _d="write",
        buf="bytes:data",
        _r="int:nbytes written",
    )

    async def cmd_w(self, buf: bytes) -> int:
        """
        Write to SPI.

        Args:
            buf: data to write
        """

        async def _run():
            await to_thread(self._bus.write, buf)
            return len(buf)

        return await self._with_cs(_run)

    doc_rw = dict(
        _d="read+write simultaneous",
        _0="bytes:data buffer (read into same buffer)",
        _r="bytes:read result",
        n="int:bytes to fill (>0) / chop off the end (<0)",
        wr="int:fill byte for n>0",
    )

    async def cmd_rw(self, buf: bytes, n: int = 0, wr: int = 0) -> bytes:
        """
        Simultaneous write and read (full duplex).

        Writes buf while reading into the same buffer.

        Args:
            buf: data to write
            n: if >0, add n fill bytes to the end; if <0, chop -n bytes off
               the result before returning it
            wr: the byte to use as filler for n>0

        """

        async def _run():
            if n > 0:
                buf += bytes((wr,)) * n
            await to_thread(self._bus.write_readinto, buf, buf)
            if n < 0:
                buf = memoryview(buf)[:-n]
            return buf

        return await self._with_cs(_run)



"""
Test module for proxying
"""

from __future__ import annotations

from moat.lib.proxy import as_proxy

# ruff: noqa: D101, D103


class Bar:
    def __init__(self, x):
        self.x = x

    def __repr__(self):
        return f"{self.__class__.__name__}.x={self.x}"


@as_proxy("fu")
class Foo(Bar):
    pass


b = as_proxy("b", Bar(95))


async def run():
    pass



"""
Console data handling for stream layers.
"""

from __future__ import annotations

from moat.lib.micro import Event

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import MutBuffer


class _CReader:
    """
    A mix-in that processes incoming console data.
    """

    def __init__(self, cons: bool | int):
        if cons is True:
            try:
                __import__("machine")
            except ImportError:
                cons = 32768
            else:
                cons = 240
        self.cevt = Event()
        self.cpos = 0
        self.cbuf = bytearray(cons)
        self.cons = cons
        self.intr = 0

    async def crd(self, buf: MutBuffer) -> int:
        """read waiting console data"""
        if not self.cons:
            raise EOFError
        if not self.cpos:
            await self.cevt.wait()
            self.cevt = Event()
        n = min(len(buf), self.cpos)
        buf[:n] = self.cbuf[:n]
        if n < self.cpos:
            self.cbuf[: self.cpos - n] = self.cbuf[n : self.cpos]
            self.cpos -= n
        else:
            self.cpos = 0
        return n

    def cput(self, b: int):
        """store a byte in the console buffer"""
        if self.cpos == len(self.cbuf):
            if len(self.cbuf) > 100:
                bfull = b"\n?BUF\n"
                self.cbuf[0 : len(bfull)] = bfull
                self.cpos = len(bfull)
            else:
                self.cpos = 0
        if b != 3:
            self.intr = 0
        elif self.intr > 2:
            raise KeyboardInterrupt
        else:
            self.intr += 1
        self.cbuf[self.cpos] = b
        self.cpos += 1
        self.cevt.set()



"""
Link RTC backend.

Provides remote access to RTC data via the root command handler.
"""

from __future__ import annotations

from . import RTC as _RTC
from . import NotSync, RTCBase

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import builtins

    from typing import Any


class RTC(RTCBase):
    """
    Link RTC backend.

    Provides remote access to RTC data via self.root.sub_at(self.cfg["path"]).
    This backend is asynchronous and cannot be used in synchronous contexts.

    The backend is expected to be callable with zero to two arguments::

    - *No argument*: return a list of available names.
    - name(str): return the data stored under that name.
    - name(str), `NotGiven`: delete the data stored under that name.
    - name(str), data(Any): store the data under that name.
    """

    is_FS = False
    is_ASYNC = True

    def __init__(self, cfg):
        super().__init__(cfg)
        self._path = cfg.get("path", ("rtc",))
        if isinstance(self._path, str):
            self._path = tuple(self._path.split("."))

    def _get_sender(self):
        """Get the SubMsgSender for our path."""
        root = _RTC.root
        if root is None:
            raise NotSync("No root connection")
        return root.sub_at(self._path)

    def reload(self):
        """Reload is a no-op for the link backend."""
        pass

    async def get(self, name: str) -> Any:
        """
        Get a value from the remote RTC.

        Args:
            name: The key to retrieve.
            all: Unused, for API compatibility.

        Returns:
            The stored value.

        Raises:
            KeyError: If the key doesn't exist.
            NotSync: If no root connection is available.
        """
        sender = self._get_sender()
        try:
            return await sender(name)
        except KeyError:
            raise
        except Exception as exc:
            raise KeyError(name) from exc

    async def set(self, name: str, data: Any) -> bool:
        """
        Set a value in the remote RTC.

        Args:
            name: The key to store.
            data: The value to store. `NotGiven` deletes.

        Returns:
            True if the value was written.

        Raises:
            NotSync: If no root connection is available.
        """
        sender = self._get_sender()
        await sender(name, data)
        return True

    async def keys(self) -> builtins.set[str]:
        """
        Get all stored data from the remote.

        Returns:
            A set of all stored values.

        Raises:
            NotSync: If no root connection is available.
        """
        sender = self._get_sender()
        return set(await sender())



"OS Error map"

from __future__ import annotations

os_error_mapping = {
    1: ("EPERM", "Operation not permitted"),
    2: ("ENOENT", "No such file or directory"),
    3: ("ESRCH", "No such process"),
    4: ("EINTR", "Interrupted system call"),
    5: ("EIO", "I/O error"),
    6: ("ENXIO", "No such device or address"),
    7: ("E2BIG", "Argument list too long"),
    8: ("ENOEXEC", "Exec format error"),
    9: ("EBADF", "Bad file number"),
    10: ("ECHILD", "No child processes"),
    11: ("EAGAIN", "Try again"),
    12: ("ENOMEM", "Out of memory"),
    13: ("EACCES", "Permission denied"),
    14: ("EFAULT", "Bad address"),
    15: ("ENOTBLK", "Block device required"),
    16: ("EBUSY", "Device or resource busy"),
    17: ("EEXIST", "File exists"),
    18: ("EXDEV", "Cross-device link"),
    19: ("ENODEV", "No such device"),
    20: ("ENOTDIR", "Not a directory"),
    21: ("EISDIR", "Is a directory"),
    22: ("EINVAL", "Invalid argument"),
    23: ("ENFILE", "File table overflow"),
    24: ("EMFILE", "Too many open files"),
    25: ("ENOTTY", "Not a typewriter"),
    26: ("ETXTBSY", "Text file busy"),
    27: ("EFBIG", "File too large"),
    28: ("ENOSPC", "No space left on device"),
    29: ("ESPIPE", "Illegal seek"),
    30: ("EROFS", "Read-only file system"),
    31: ("EMLINK", "Too many links"),
    32: ("EPIPE", "Broken pipe"),
    33: ("EDOM", "Math argument out of domain of func"),
    34: ("ERANGE", "Math result not representable"),
    # E'EWOULDBLOCK', AGAIN: ('operation would block'),
    95: ("EOPNOTSUPP", "Operation not supported on transport endpoint"),
    97: ("EAFNOSUPPORT", "Address family not supported by protocol"),
    98: ("EADDRINUSE", "Address already in use"),
    103: ("ECONNABORTED", "Software caused connection abort"),
    104: ("ECONNRESET", "Connection reset by peer"),
    105: ("ENOBUFS", "No buffer space available"),
    106: ("EISCONN", "Transport endpoint is already connected"),
    107: ("ENOTCONN", "Transport endpoint is not connected"),
    110: ("ETIMEDOUT", "Connection timed out"),
    111: ("ECONNREFUSED", "Connection refused"),
    113: ("EHOSTUNREACH", "No route to host"),
    114: ("EALREADY", "Operation already in progress"),
    115: ("EINPROGRESS", "Operation now in progress"),
}



"""
Data transfer handler.

This is intended for regularly moving data from A to B.
"""

from __future__ import annotations

from moat.util import Queue, QueueFull, combine_dict
from moat.lib.micro import Event, L, TaskGroup, every_ms, idle, is_async, log, ticks_ms
from moat.lib.path import Path
from moat.lib.rpc import BaseCmd, RemoteError

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.path import PathElem
    from moat.lib.rpc import Msg, SubMsgSender

    from collections.abc import Mapping, Sequence


class _Send:
    # A null context that delegates its .send method to the wrapped destination
    def __init__(self, dest):
        self.dest = dest

    def send(self, *a, **kw):
        return self.dest(*a, **kw)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        return None


class _Step:
    """
    One transfer step.

    A step has an input (it's called, async) and an output (its set of queues).
    The `si` and `so` flags control streaming, viewed *as by the app*, i.e.
    "stream in" means "we send data to it in a stream"

    Four cases:
    - si=so=False:
      simple-call the path with incoming data
      the result is the data coming back.
      A timer or an incoming call starts the chain.
    - si=True
      Open a stream to send incoming data.
      A timer calls `cmd`.
    - so=True
      Open a stream to receive data.
      Input is sent to `cmd`, or lost. No timer.
    - si=so=True
      Input is sent to the stream and output is received from it.
      A timer may periodically run `cmd`.
    """

    call = None  # subcmd or feeder

    p: SubMsgSender | None = None
    a: tuple
    k: dict
    append: bool = False
    keep: bool = False

    t: int | None = None
    si: bool = False
    so: bool = False
    msg: Msg | None = None

    last_a: Sequence | None = None
    last_kw: Mapping | None = None
    _ready: Event

    def __init__(self, trans: Transfer, id: int, cfg: Path | dict):
        self.trans = trans
        self.id = id
        self.q = set()  # output queue(s)
        self._ready = Event()
        self.is_ready = self._ready.wait

        if isinstance(cfg, Path):
            p = cfg
            self.a = ()
            self.kw = {}
        else:
            p = cfg.get("p", None)
            self.a = tuple(cfg.get("a", ()))
            self.kw = cfg.get("k", {})
            self.si = cfg.get("si", self.si)
            self.so = cfg.get("so", self.so)
            self.append = cfg.get("append", self.append)
            self.keep = cfg.get("keep", self.keep)
        if p is not None:
            sp = trans.root.sub_at(p)
            if not hasattr(sp, "stream_out"):
                raise TypeError(sp)
            self.p = sp

    def __repr__(self):
        return f"‹TS:{self.trans.path / self.id}›"

    async def run(self) -> None:
        if self.p is not None:
            await self.p.rdy_()
        if self.so:
            await self.run_o()
        elif self.si:
            await self.run_i()
        else:
            await self.run_t()

    async def run_i(self) -> None:
        "Open a write-only pipe to the remote"
        assert self.p is not None
        async with self.p.stream_out(*self.a, **self.kw) as self.msg:
            self._ready.set()
            await self.run_t()

    async def run_o(self) -> None:
        "Open a pipe to the remote and read from it"
        assert self.p is not None
        async with (self.p.stream if self.si else self.p.stream_in)(*self.a, **self.kw) as msg:
            self._ready.set()
            if self.si:
                self.msg = msg
            async for m in msg:
                await self.cont(m.args, m.kw)

    async def run_t(self) -> None:
        "Run the task."
        self._ready.set()
        await idle()

    async def __call__(self, akw: tuple[list, dict] | None = None) -> tuple[list, dict] | None:
        """
        The previous step delivers data to us.
        """
        if akw is None:
            a = ()
            kw = {}
        else:
            a, kw = akw
        if self.keep:
            ra, rkw = a, kw
        if self.si:
            await self._ready.wait()
            assert self.msg is not None
            await self.msg.send(*a, **kw)
            return

        if self.so:
            # nothing to do
            log(f"Dropped {a} {kw}")
            pass

        if self.p is not None:
            if self.a:
                if self.append:
                    a = self.a + tuple(a)
                else:
                    a = tuple(a) + self.a
            if self.kw:
                kw = combine_dict(kw, self.kw)
            try:
                msg = await self.p.cmd(Path(), *a, **kw)
            except Exception as exc:
                raise RemoteError(f"{self}: {exc}") from exc

            self.last_a = msg.args
            self.last_kw = msg.kw

            if self.keep:
                a = ra
                kw = rkw
            else:
                a = msg.args
                kw = msg.kw

        # pass on
        await self.cont(a, kw)

    async def cont(self, a, kw) -> None:
        """
        Take this data and continue to the next step
        """
        for q in self.q:
            try:
                res = q((a, kw))
                if is_async(res):
                    await res
            except QueueFull:
                pass


class Transfer(BaseCmd):
    """
    This command calls A, then calls B with the result, and so on.

    Config:
    - t: frequency at which to call A.
    - s: an array of calls.
         Either a path, or a map with
      - p: the path to use
      - a: positional arguments
      - k: keyword arguments
      - append: flag to add values at the end of a instead of in front
      - keep: flag to not replace values
      - si: Stream-in; bool, defaults to False.
      - so: Stream-out; bool, defaults to False.

      TODO: we might want a step for argument mangling

      - put: path (or list of paths) to insert data into the arguments
      - get: path (or list of paths) to extract data from the result
    """

    steps: list[_Step]
    t_last: int | None = None

    doc = dict(
        _c=dict(
            _d="Data transfer",
            t="int:delay(ms)",
            _a=[
                dict(s="path:called"),
                dict(
                    s=dict(
                        p="path:called",
                        a="any:pos args",
                        k="any:kw args",
                        append="bool:at end?",
                        keep="bool:out only",
                        si="bool:stream in?",
                        so="bool:stream out?",
                    )
                ),
            ],
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        self.steps = []
        self.data = []

    async def task(self):
        "Step tasks runner"
        while True:
            # TG gets cancelled when reloading
            async with TaskGroup() as self.tg:
                for i, step in enumerate(self.cfg["s"]):
                    s = _Step(self, i, step)
                    if self.steps:
                        self.steps[-1].q.add(s)
                    self.steps.append(s)

                for s in self.steps[::-1]:
                    self.tg.start_soon(s.run)
                for s in self.steps:
                    await s.is_ready()
                if L:
                    self.set_ready()

                t = self.cfg.get("t", None)
                if t is None:
                    await idle()
                else:
                    async for _ in every_ms(t):
                        self.t_last = ticks_ms()
                        await self.steps[0]()

    async def reload(self):
        "restart"
        await super().reload()
        self.tg.cancel()

    async def cont(self, i, a, kw):
        "Data forwarding"
        self.data[i] = (a, kw)
        i += 1
        if i < len(self.data):
            s = self.steps[i]
            if not s.si:
                await s((a, kw))

    doc_w = dict(_d="data", qs="int:dest step (default first)", _s=True)

    async def stream_w(self, msg: Msg):
        "send data to (first) step"
        kw = msg.kw
        s0 = self.steps[kw.pop("qs", 0)]
        if msg.can_stream:
            async with msg.stream_in() as md:
                async for m in md:
                    await s0((list(m.args), m.kw))
        else:
            await s0((list(msg.args), kw))

    doc_r = dict(
        _d="data monitor",
        qs="int:src step (last)",
        cur="bool:get current value if present",
        _s=True,
    )

    async def stream_r(self, msg: Msg):
        "read data from (last) step"
        s0 = self.steps[msg.get("qs", -1)]
        q = Queue(1)
        qp = q.put
        s0.q.add(qp)
        try:
            if msg.can_stream:
                async with msg.stream_out(*(s0.last_a or ()), **(s0.last_kw or {})) as md:
                    if s0.last_a is not None and msg.get("cur", False):
                        await md.send(*s0.last_a, **s0.last_kw)
                    async for a, kw in q:
                        await md.send(*a, **kw)
            else:
                if s0.last_a is not None and msg.get("cur", False):
                    a, kw = s0.last_a, s0.last_kw
                else:
                    a, kw = await q.get()
                await msg.result(*a, **kw)
        finally:
            s0.q.remove(qp)

    doc_s = dict(_d="send state", _o="any")

    async def stream_s(self, msg: Msg):
        "feed current state"
        async with msg.stream_out(t=self.t_last) as md:
            for a, kw in self.data:
                await md.send(*a, **kw)

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> None:
        "Intercept qs"
        if rcmd and isinstance(rcmd[-1], int):
            msg.kw["qs"] = rcmd.pop()
        await super().handle(msg, rcmd)



"""
Basic command handling in MoaT-RPC.
"""

from __future__ import annotations

from ._base import BaseCmd as BaseCmd
from ._base import LoadCmd as LoadCmd
from ._base import LockBaseCmd as LockBaseCmd
from ._base import RootCmd as RootCmd



"""
Message streaming.
"""

from __future__ import annotations

from functools import partial

from moat.util import QueueFull, pop_kw, push_kw
from moat.lib.micro import (
    ACM,
    AC_exit,
    Event,
    L,
    Queue,
    TaskGroup,
    log,
    log_exc,
    shield,
    sleep_ms,
    ticks_diff,
    ticks_ms,
)
from moat.lib.path import Path
from moat.lib.rpc import (
    B_ERROR,
    B_STREAM,
    B_WARNING,
    B_WARNING_INTERNAL,
    E_CANCEL,
    E_ERROR,
    E_NO_CMD,
    E_NO_CMDS,
    E_NO_STREAM,
    E_SKIP,
    E_UNSPEC,
    Msg,
    MsgHandler,
    MsgLink,
    ShortCommandError,
)

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.lib.micro import _TaskGroupProto
    from moat.lib.path import PathElem
    from moat.lib.rpc import BaseMsgHandler, OptDict

    from collections.abc import Callable, Sequence
    from typing import Any

__all__ = ["HandlerStream", "StreamLink", "i_f2wire", "wire2i_f"]


def B_FL_NAME(flag):
    "stringify message flags"
    if flag & B_ERROR:
        return ".W" if flag & B_STREAM else ".E"
    else:
        return ".S" if flag & B_STREAM else ""


def B_ERR_NAME(err):
    "stringify message errors"
    if not isinstance(err, int):
        return repr(err)
    if err >= 0:
        return f"S+{err}"
    if err <= E_NO_CMD:
        return f"NO_CMD_{-E_NO_CMD - err}"
    if err == E_UNSPEC:
        return "UNSPEC"
    if err == E_NO_STREAM:
        return "NO_STREAM"
    if err == E_CANCEL:
        return "CANCEL"
    if err == E_NO_CMDS:
        return "NO_CMDS"
    if err == E_SKIP:
        return "SKIP"
    if err == E_ERROR:
        return "ERROR"
    return f"?{err}"


def i_f2wire(id: int, flag: int) -> int:  # noqa: D103
    assert id != 0
    assert 0 <= flag <= 3 or flag == B_WARNING_INTERNAL
    if id > 0:
        id -= 1
    return (id << 2) | (flag & 3)


def wire2i_f(id: int) -> tuple[int, int]:  # noqa: D103
    f = id & 3
    id >>= 2
    if id >= 0:
        id += 1
    return id, f


class HandlerStream(MsgHandler):
    """
    This class bidirectionally translates MsgHandler calls to streamed messages.

    This is a sans-I/O class. Usage:

    * open an async context on an instance of this class
    * start a task that reads your external source and feeds the result
      to :meth:`~HandlerStream.msg_in`.
    * start a task that loops on :meth:`~HandlerStream.msg_out` and sends the result. It should
      terminate on `EOFError` from :meth:`~HandlerStream.msg_out`.

    You can use the :meth:`~HandlerStream.start` method to run these tasks (and any others you
    might need) within the context's internal taskgroup. They will be
    auto-cancelled when leaving the context.

    Message encoding:

    This class encodes messages as plain Python lists. They consist of
    integers plus any other data type your payload consists of.

    Error handling uses proxies or strings for the error name, but it will
    send a plain error indication if it can't encode its data. If your
    messages contain keywords, the codec needs to support dicts.

    """

    _tg: _TaskGroupProto | None = None
    _id = 0

    def __init__(self, sender: BaseMsgHandler | None, logger: object | None = None):
        self._msgs: dict[int, StreamLink] = {}
        self._send_q = Queue(9)
        self._recv_q = Queue(99)
        self._sender = sender

        self._logger = cast("Callable[..., object] | None", getattr(logger, "debug", logger))

        self.reader_done = Event()
        self.writer_done = Event()
        self.closing = True

        self._id1: set[int] = set()
        if L:
            self._id2: set[int] = set()
        self._id3: set[int] = set()

    @property
    def is_idle(self) -> bool:  # noqa: D102
        if self._msgs:
            return False
        if self._send_q.qsize():
            return False
        if self._recv_q.qsize():
            return False
        return True

    async def handle(self, msg: Msg, rcmd: list[PathElem]) -> None:
        """
        Forward a new message to the other side.
        """
        if self.closing:
            raise EOFError

        if not rcmd:
            raise ShortCommandError(msg.cmd)
        if rcmd[-1] == "!":
            rcmd.pop()
            return await super().handle(msg, rcmd)

        rcmd = rcmd[:]
        rcmd.reverse()

        i = self._gen_id()
        can_stream = msg.can_stream
        link = StreamLink(self, i)
        msg.replace_with(link)
        self.attach(link)
        args = [rcmd]
        args.extend(msg.args)
        await self.send(link, args, msg._kw, B_STREAM if can_stream else 0)  # noqa: SLF001
        if not can_stream:
            msg.set_end()
        # breakpoint()  # fwd

    def _gen_id(self) -> int:
        # Generate the next free ID.
        if self._id1:
            return self._id1.pop()
        if L and self._id2:
            return self._id2.pop()
        if self._id3:
            return self._id3.pop()
        self._id += 1
        return self._id

    async def closed_input(self) -> None:
        """
        Input is down, no more messages
        """
        self.closing = True

        if (tg := self._tg) is not None:
            self._tg = None
            self._send_q.close_sender()
            await self.writer_done.wait()
            self._recv_q.close_sender()
            if self._read_task is not None:
                self._read_task.cancel()
                await self.reader_done.wait()

            tg.cancel()

        for link in list(self._msgs.values()):
            await link.kill()

    async def msg_in(self, msg: list) -> None:
        """process an incoming message"""
        try:
            i, flag = wire2i_f(msg[0])
        except TypeError:
            raise ValueError(f"{msg!r} is {type(msg)}. Codec mismatch?") from None
        # incoming: flip the sign
        i = -i

        if self._logger:
            self._logger(
                "IN : %d%s %s%r",
                i,
                B_FL_NAME(flag),
                B_ERR_NAME(msg[1]) + " " if flag & B_ERROR else "",
                msg[2:] if flag & B_ERROR else msg[1:],
            )

        a = msg[1:]
        internal = flag == B_WARNING and len(a) == 1 and isinstance(a[0], int)
        kw = None if internal else pop_kw(a)

        stream = flag & B_STREAM
        error = flag & B_ERROR
        if internal and stream and error:
            flag = B_WARNING_INTERNAL

        try:
            link = self._msgs[i]

        except KeyError:
            if i > 0:
                log("Spurious message %r", msg)
            elif error:
                log("Spurious error %r", msg)
            else:
                # assemble the message
                cmd = a.pop(0) if a else Path()
                rem = Msg.Call(cmd, a, {} if kw is None else kw, flag)

                # … and build a stream for it
                link = StreamLink(self, i)
                rem.replace_with(link)
                if not stream:
                    link.set_end()
                self.attach(link)
                link.task = await self._tgs.spawn(self._handle, msg, link)
        else:
            try:
                await link.ml_send(a, kw, flag)
            except EOFError:
                try:
                    await self.send(link, [E_NO_STREAM], None, B_ERROR)
                except EOFError:
                    pass
                self.detach(link)
            except QueueFull:
                if flag & B_STREAM:
                    await self.send(link, [E_SKIP], None, B_WARNING_INTERNAL)
                else:
                    await self.send(link, [E_SKIP], None, B_ERROR)
                    self.detach(link)

            else:
                if link.end_both:
                    self.detach(link)

    async def _handle(self, msg: list, link: StreamLink) -> None:
        """
        Task for a new incoming connection.
        """
        if self._sender is None:
            await self.send(link, [E_NO_CMD], None, B_ERROR)
            return
        if self.closing:
            if not link.end_here:
                await self.send(link, [E_CANCEL], None, B_ERROR)
            return

        rem = cast(Msg, link.remote)
        try:
            res = await self._sender.handle(rem, rem.rcmd)
            if res is not None:
                if link.end_there:
                    raise ValueError(f"Already ended but returned {res!r}")
                else:
                    await link.remote.ml_send([res], None, 0)
        except Exception as exc:
            log_exc(exc, "Error %r: %r", msg, exc)
            if link.remote is not None:
                await link.remote.ml_send_error(exc)
        except BaseException as exc:
            if link.remote is not None:
                with shield():
                    await link.remote.ml_send_error(exc)
            raise
        else:
            # may have been replaced by the handler
            if rem is link.remote and not rem.end_here:
                await self.send(link, [None], None, 0)

        if not link.end_both:
            # log("NotClosed L%d L%d", link.link_id, link.remote.link_id if link.remote else -1)
            # breakpoint() # notclosed
            pass  # raise RuntimeError("Link was not closed")

    def attach(self, proc: StreamLink) -> None:
        """
        Attach a link.
        """
        if proc.id in self._msgs:
            raise ValueError(f"MID {proc.id} already known")
        self._msgs[proc.id] = proc

    def detach(self, link: StreamLink) -> None:
        """
        Remove a link.
        """
        mid = link.id
        if self._msgs.get(mid) is not link:
            # already removed
            return
        del self._msgs[mid]
        if mid <= 0:  # remote
            return

        # Optimizing for CBOR integers
        if L:
            self._dly_q.put_nowait((mid, ticks_ms()))
        else:
            if mid < 6:
                self._id1.add(mid)
            else:
                self._id3.add(mid)

    async def _dly(self):
        "Delay for message IDs because we don't want to re-use them immediately."
        while True:
            mid, t = await self._dly_q.get()
            tt = ticks_ms()
            td = ticks_diff(tt, t)
            if td < 1000:
                await sleep_ms(1000 - td)
            if mid < 6:
                self._id1.add(mid)
            elif mid < 64:
                self._id2.add(mid)
            else:
                self._id3.add(mid)

    async def send(self, link: StreamLink, a: Sequence, kw: OptDict, flag: int) -> None:  # noqa: D102
        if self.closing:
            raise EOFError
        assert isinstance(a, (list, tuple)), a
        assert 0 <= flag <= 3 or flag == B_WARNING_INTERNAL, flag
        # log("SendQ L%d %r %r %d", link.link_id, a, kw, flag)
        await self._send_q.put((link, a, kw, flag))

    async def msg_out(self) -> list:
        """
        Get the next outgoing message to be sent over the stream.

        Returns:
            A list representing the encoded message ready to be sent.
        """
        link, a, kw, flag = await self._send_q.get()
        i = i_f2wire(link.id, flag)

        res: list[Any] = [i]
        res.extend(a)
        push_kw(res, kw, is_warn=(flag == B_WARNING))

        if self._logger:
            self._logger(
                "OUT: %d%s %s%r",
                link.id,
                B_FL_NAME(flag),
                B_ERR_NAME(res[1]) + " " if flag & B_ERROR else "",
                res[2:] if flag & B_ERROR else res[1:],
            )
        return res

    def start(self, cmd, *a, **kw) -> None:
        """
        Start a task within the context's internal taskgroup.

        Args:
            cmd: The coroutine function to run.
            *a: Positional arguments to pass to the function.
            **kw: Keyword arguments to pass to the function.
        """
        if self._tg is None:
            raise RuntimeError("Not started")
        if kw:
            self._tg.start_soon(partial(cmd, *a, *kw))
        else:
            self._tg.start_soon(cmd, *a)

    async def __aenter__(self):
        acm = ACM(self)
        try:
            tg: _TaskGroupProto = await acm(TaskGroup())
            self._tg = tg
            self._tgs = await acm(TaskGroup())
            self.closing = False

            evt1 = Event()
            evt2 = Event()
            self._read_task = await tg.spawn(self._run_read, evt1)
            self._write_task = await tg.spawn(self._run_write, evt2)
            await evt1.wait()
            await evt2.wait()
            if L:
                self._dly_q = Queue(999)
                await acm(self._dly_q.close_sender)
                tg.start_soon(self._dly)
            return self

        except BaseException as exc:
            await AC_exit(self, type(exc), exc, None)
            raise

    async def _run_read(self, evt):
        try:
            if self.reader_done.is_set():
                self.reader_done = Event()
            evt.set()
            await self.read_stream()
        finally:
            self.reader_done.set()
            self._read_task = None

    async def _run_write(self, evt):
        try:
            if self.writer_done.is_set():
                self.writer_done = Event()
            evt.set()
            await self.write_stream()
        finally:
            self.writer_done.set()
            self._write_task = None

    async def read_stream(self):
        """
        Stream reader task.

        *Must be overridden*: Iterate calling :meth:`~HandlerStream.msg_out`
        and writing its return value to your stream.
        """
        raise NotImplementedError

    async def write_stream(self):
        """
        Stream writer task.

        *Must be overridden*: Iterate reading from your stream and calling
        :meth:`~HandlerStream.msg_in` with the result.
        """
        raise NotImplementedError

    async def __aexit__(self, *exc):
        self._tgs.cancel()
        try:
            with shield():
                await self.closed_input()
            self._recv_q.close_sender()
            self._msgs = {}

            with shield():
                for link in list(self._msgs.values()):
                    await link.kill()
            assert not self._msgs

        finally:
            await AC_exit(self, *exc)


class StreamLink(MsgLink):
    """This is the handler for messages that forwards them to the stream."""

    def __init__(self, stream: HandlerStream, id: int):
        super().__init__()
        self.__stream: HandlerStream | None = stream
        self.id = id
        self.task = None

    async def ml_recv(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """data to be forwarded across the link"""
        if self.__stream is None:
            raise EOFError
        assert 0 <= flags <= 3 or flags == B_WARNING_INTERNAL, flags
        # log("LR L%d %d %r %r %d", self.link_id, self.id, a, kw, flags)
        await self.__stream.send(self, a, kw, flags)

    async def ml_send(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """data to be forwarded to our remote"""
        # log("LS L%d %d %r %r %d", self.link_id, self.id, a, kw, flags)
        assert 0 <= flags <= 3 or flags == B_WARNING_INTERNAL, flags
        await super().ml_send(a, kw, flags)

    def stream_detach(self) -> None:  # noqa: D102
        if self.__stream is not None:
            self.__stream.detach(self)
            self.__stream = None
        if self.task is not None:
            try:
                self.task.cancel()
            except RuntimeError:
                pass  # self-cancel on µPy is forbidden
            else:
                self.task = None



"""
This module contains list-to-kw mangling helpers.
"""

from __future__ import annotations

from moat.lib.micro import Mapping

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from typing import Any

__all__ = ["pop_kw", "push_kw"]


def push_kw(args: list[Any], kwargs: dict[str, Any], is_warn: bool = False) -> None:
    """
    Add kwargs to the list, if required.

    Args:
        args:
            The argument list.
        kwargs:
            The argument mapping to (possibly) append.
        is_warn:
            This message is a user-level warning.

    The mapping is appended if
    - it is not empty.
    - the last argument is a mapping.
    - the argument of a user-level warning consists of a single integer.
    """
    if (
        kwargs
        or (args and isinstance(args[-1], dict))
        or (is_warn and len(args) == 1 and isinstance(args[0], int))
    ):
        args.append(kwargs if isinstance(kwargs, dict) else {})


def pop_kw(ak: list[Any]) -> MutableMapping[str, Any]:
    """
    Unpack an args-and-maybe-trailing-kwargs list.

    This function removes (and returns) an argument lists's trailing mapping, if any.
    Otherwise the result is an empty dict.

    Args:
        ak: Argument list, as generated by `push_kw`:

    Returns:
        The trailing mapping, or an empty dict.
    """
    if ak and isinstance(ak[-1], Mapping):
        return ak.pop()
    return {}



"""
File system RTC backend.

Stores data in a CBOR file on the file system.
Supports direct file mapping for specific keys.
"""

from __future__ import annotations

from moat.util import NotGiven
from moat.lib.codec.cbor import CBOR_TAG_CBOR_FILEHEADER, Tag
from moat.lib.codec.moat_cbor import Codec as CBOR
from moat.lib.micro import log

from . import RTCBase

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import builtins

_codec = CBOR()


class RTC(RTCBase):
    """
    File system RTC backend.

    Stores data in a CBOR file at cfg["dest"] (default: "moat.data").
    Supports direct file mapping via cfg["direct"] (key→filename dict):
    values for these keys must be strings and are written directly to files.
    """

    is_FS = True
    is_ASYNC = None  # configurable

    def __init__(self, cfg):
        super().__init__(cfg)
        self._d = {}
        self._dest = cfg.get("dest", "moat.data")
        self._direct = cfg.get("direct", {"state": "moat.state"})
        self._load()

    def _load(self):
        """Load data from the state file."""
        try:
            with open(self._dest, "rb") as f:
                data = f.read()
            if data:
                decoded = _codec.decode(data)
                if isinstance(decoded, Tag) and decoded.tag == CBOR_TAG_CBOR_FILEHEADER:
                    self._d = decoded.value
                elif isinstance(decoded, dict):
                    self._d = decoded
        except OSError:
            # File doesn't exist yet
            self._d = {}
        except Exception as exc:
            log("RTC fs load error", err=exc)
            self._d = {}

    def _save(self):
        """Save data to the state file."""
        tagged = Tag(CBOR_TAG_CBOR_FILEHEADER, self._d)
        with open(self._dest, "wb") as f:
            f.write(_codec.encode(tagged))

    def _get_direct(self, name: str):
        """Get a value from a direct file."""
        fn = self._direct[name]
        try:
            with open(fn, "r") as f:
                return f.read()
        except OSError:
            raise KeyError(name) from None

    def _set_direct(self, name: str, data: str) -> bool:
        """Set a value in a direct file."""
        fn = self._direct[name]
        with open(fn, "w") as f:
            f.write(data)
        return True

    def _del_direct(self, name: str) -> bool:
        """Delete a direct file."""
        fn = self._direct[name]
        try:
            import os  # noqa:PLC0415

            os.remove(fn)
        except OSError:
            pass
        return True

    def reload(self):
        """Reload data from the state file."""
        self._load()

    async def get(self, name: str):
        """
        Get a value from the file system.

        Args:
            name: The key to retrieve.
            all: Unused, for API compatibility.

        Returns:
            The stored value.

        Raises:
            KeyError: If the key doesn't exist.
        """
        # Check direct files first
        if name in self._direct:
            return self._get_direct(name)

        if name not in self._d:
            raise KeyError(name)
        return self._d[name]

    async def set(self, name: str, data) -> bool:
        """
        Set a value in the file system.

        Args:
            name: The key to store.
            data: The value to store.

        Returns:
            True if the value was written.
        """
        # Handle direct files
        if name in self._direct:
            if data is NotGiven:
                return self._del_direct(name)
            if not isinstance(data, str):
                raise TypeError(f"Direct file {name} requires string value")
            try:
                if self._get_direct(name) == data:
                    return True
            except KeyError:
                pass
            return self._set_direct(name, data)

        if data is NotGiven:
            if name not in self._d:
                return True
            del self._d[name]
        else:
            self._d[name] = data
        self._save()
        return True

    async def keys(self) -> builtins.set[str]:
        """
        Get all available keys.

        Returns:
            A copy of the internal data dict (does not include direct files).
        """
        res = set(self._direct.keys())
        res.update(self._d.keys())
        return res



"""
Packetized serial port app.
"""

from __future__ import annotations

from moat.util import attrdict
from moat.lib.micro import AC_use
from moat.lib.rpc import BaseCmdBBM
from moat.lib.stream import SerialPackerBlkBuf
from moat.micro.app._doc import _cons_d, _frame_d, _mode_d

from ._util import get_serial


class Msg(BaseCmdBBM):
    """snd/rcv: packetized data, via SerialPacker."""

    doc = dict(
        _c=dict(
            _d="packet serial data",
            port="str:Port or 'USB'",
            mode=_mode_d,
            frame=_frame_d,
            cons=_cons_d,
        )
    )

    async def stream(self):
        """Returns the packetized serial stream."""
        ser = SerialPackerBlkBuf(
            get_serial(self.cfg),
            frame=self.cfg.get("frame", attrdict()),
            console=self.cfg.get("cons", False),
        )
        return await AC_use(self, BaseCmdBBM(ser))



from __future__ import annotations

import gc
import sys

import machine

from moat.lib.micro import log, sleep_ms, ticks_diff, ticks_ms, to_thread

from ._sys_ import Cmd as _Cmd


class Cmd(_Cmd):
    """
    System stuff that's satellite specific
    """

    doc_state = dict(_d="Root info")

    async def cmd_state(self):
        """
        Return the root info.
        """
        return self.root.i

    doc_rtc = dict(_d="RTC state access", _0="str:Key", fs="bool:filesystem", v="Any:value to set")

    async def cmd_rtc(self, k="state", v=None, fs=None):
        """
        Set/return a MoaT state var.
        """
        from moat.micro.rtc import RTC  # noqa: PLC0415

        if v is not None:
            RTC.set_sync(k, v, fs=fs)
        else:
            return RTC.get_sync(k, fs=fs, default=None)

    doc_mem = dict(
        _d="memory info",
        _r=dict(
            c="int:bytes freed",
            t="int:gc run(ms)",
            a=["int:alloc now", "int:alloc at boot"],
            f=["int:free now", "int:free at boot"],
        ),
    )

    async def cmd_mem(self):
        """
        Info about memory. Calls `gc.collect`.

        * c: bytes freed by the garbage collector
        * t: time (ms) for the garbage collector to run
        * a: allocation: (now,early)
        * f: free memory: (now,early)
        """
        t1 = ticks_ms()
        f1 = gc.mem_free()
        gc.collect()
        f2 = gc.mem_free()
        a2 = gc.mem_alloc()
        t2 = ticks_ms()
        return dict(
            t=ticks_diff(t2, t1),
            a=(a2, self.root.i["fa"]),
            f=(f2, self.root.i["fm"]),
            c=f2 - f1,
        )

    doc_hash = dict(_d="module hash", _0="str:name", l="int:shortened length", _r="str:hash val")

    async def cmd_hash(self, p: str, l: int | None = None):  # noqa:E741
        """
        Get the hash for a built-in module.
        """
        import _hash  # noqa: PLC0415

        res = _hash.hash.get(p, None)
        if l is not None:
            res = res[:l]
        return res

    doc_log = dict(_d="call log", _0="str:text", _99="any:params", _k="any:params")

    async def cmd_log(self, *a, **k):
        """
        Log parameters.
        """
        log(f"Input: {a!r} {k!r}")

    doc_stdout = dict(_d="write stdout", _0="str:text", _99="any:params", _k="any:params")

    async def cmd_stdout(self, *a, **k):
        """
        Print something to stdout.

        WARNING this may disrupt communication if stdout is re-used for the MoaT link.
        """
        print(f"Input: {a!r} {k!r}", file=sys.stderr)

    doc_stderr = dict(_d="write stderr", _0="str:text", _99="any:params", _k="any:params")

    async def cmd_stderr(self, *a, **k):
        """
        Print something to stderr.
        """
        print(f"Input: {a!r} {k!r}", file=sys.stderr)

    doc_boot = dict(
        _d="reboot",
        _0="str:SysBooT",
        _1="int:(0)=hard,1=soft,2=boot,3=KbdIntr,4=immediate,5=RuntimeExc",
        _t="int:timeout msec(100)",
    )

    async def cmd_boot(self, code, m=0, t=100):
        """
        Reboot MoaT.

        @code needs to be "SysBooT".

        @m can be
            0: hard reset (via SystemExit)
            1: soft reset (via SystemExit)
            2: bootloader
            3: KeyboardInterrupt
            4: hard reset (immediate)
            5: raise RuntimeError

        Modes other than 0/4 and 2 do not work (not for long anyway) when
        the watchdog module is active.
        """
        if code != "SysBooT":
            raise RuntimeError("wrong")

        async def _boot(t):
            await sleep_ms(t)
            if m == 0:
                raise SystemExit(1)
            if m == 1:
                raise SystemExit(0)
            elif m == 2:
                machine.bootloader()
            elif m == 3:
                raise KeyboardInterrupt
            elif m == 4:
                machine.reset()
            elif m == 5:
                raise RuntimeError("Boot")

        self.root.tg.start_soon(_boot, t, _name="_sys.boot1")
        return True

    doc_machid = dict(_d="unique machine id", _r="int")

    async def cmd_machid(self):
        """
        Return the machine's unique ID. This is the bytearray returned by
        `micropython.unique_id`.
        """
        return machine.unique_id()

    doc_pin = dict(
        _d="direct digital pin", _0="int:pin#", v="bool:value to set", _r="bool:hw state"
    )

    async def cmd_pin(self, n, v=None, **kw):
        """
        Set or read a digital pin.
        """
        p = machine.Pin(n, **kw)
        if v is not None:
            p.value(v)
        return p.value()

    doc_adc = dict(_d="direct analog pin", _0="int:pin#", _r="int:16-bit value")

    async def cmd_adc(self, n):
        """
        Read an analog pin.
        """
        p = machine.ADC(n)
        return await to_thread(p.read_u16)



"""
TCP link app for MoaT message exchange.
"""

from __future__ import annotations

from moat.util import attrdict
from moat.lib.rpc import CmdMsg
from moat.lib.stream import TcpLink, serial_stack


class Link(CmdMsg):
    """
    An app that connects to a remote socket.
    """

    def __init__(self, cfg):
        stack = serial_stack(
            TcpLink(
                cfg.get("host", "127.0.0.1"),
                cfg["port"],
                retry=cfg.get("retry", attrdict()),
            ),
            cfg,
        )
        super().__init__(cfg, stack)



"""
Packet handler for diyBMS-serial mock battery
"""

from __future__ import annotations

import moat.ems.battery.diy_serial.packet as P
from moat.ems.battery.errors import MessageError
from moat.micro.conv.steinhart import celsius2thermistor

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.ems.battery.diy_serial.packet import _Reply

__all__ = [
    "MAXCELLS",
    "PacketHeader",
    "PacketType",
    "replyClass",
    "requestClass",
]

PacketType = P.PacketType
MAXCELLS = P.MAXCELLS


class PacketHeader(P.PacketHeader):
    def decode_one(self, msg):
        """Decode a single message to packet + rest.

        This method is used by the client.
        """
        off = 0
        RC = requestClass[self.command]
        pkt_len = RC.S.size
        pkt = None
        if pkt_len:
            if off + pkt_len > len(msg):
                raise MessageError(bytes(msg))  # incomplete
            pkt = RC.from_bytes(msg[off : off + pkt_len])
            if not self.broadcast:
                off += pkt_len
        if off:
            msg = memoryview(msg)[off:]
        return pkt, msg

    def encode_one(self, msg, pkt: _Reply = None):
        return self.to_bytes() + msg + (pkt.to_bytes() if pkt is not None else b"")


###


class ReplyIdentify(P.ReplyIdentify):
    def to_bytes(self):
        return b""


class RequestTiming(P.RequestTiming):
    async def to_cell(self, _cell):
        _cell._dp_t = self.timer  # noqa: SLF001


class ReplyTiming(P.ReplyTiming):
    async def from_cell(self, _cell):
        self.timer = _cell._dp_t  # noqa: SLF001


class ReplyReadSettings(P.ReplyReadSettings):
    async def from_cell(self, cell):
        bc_i, bc_e = await cell.b_coeff()
        s = await cell.settings()
        self.BCoeffInternal = bc_i
        self.BCoeffExternal = bc_e
        self.gitVersion = 0
        self.boardVersion = 0
        self.dataVersion = 0
        self.mvPerADC = int(s["vpa"] * 1000 * 64)
        self.voltageCalibration = s["vcal"]
        self.bypassTempRaw = 0
        self.bypassVoltRaw = 0
        self.numSamples = s["ns"]
        self.loadResRaw = int(4.2 * 16 + 0.5)

    def to_bytes(self):
        return self.S.pack(
            self.gitVersion,
            self.boardVersion,
            self.dataVersion,
            self.mvPerADC,
            self.voltageCalibration,
            self.bypassTempRaw,
            self.bypassVoltRaw,
            self.BCoeffInternal,
            self.BCoeffExternal,
            self.numSamples,
            self.loadResRaw,
        )


class ReplyVoltages(P.ReplyVoltages):
    def to_bytes(self):
        return self.S.pack(self.voltRaw, self.bypassRaw)

    async def from_cell(self, cell):
        u = await cell.u()
        bal = await cell.bal()

        self.voltRaw = await cell.v2raw(u)
        if bal.get("th", None) is None:
            self.bypassRaw = 0
        else:
            self.bypassRaw = await cell.v2raw(bal["th"])
        if bal["b"]:
            self.voltRaw |= 0x8000
        if bal["ot"]:
            self.voltRaw |= 0x4000


class ReplyTemperature(P.ReplyTemperature):
    async def from_cell(self, cell):
        t = await cell.t()
        tb = await cell.tb()
        bc_i, bc_e = await cell.b_coeff()
        self.intRaw = celsius2thermistor(bc_i, tb)
        self.extRaw = celsius2thermistor(bc_e, t)

    def to_bytes(self):
        b1 = self.intRaw & 0xFF
        b2 = (self.intRaw >> 8) | ((self.extRaw & 0x0F) << 4)
        b3 = self.extRaw >> 4
        return self.S.pack(b1, b2, b3)


requestClass = [
    P.RequestResetCounters,  # ResetCounters=0
    P.RequestVoltages,  # ReadVoltages=1
    P.RequestIdentify,  # Identify=2
    P.RequestTemperature,  # ReadTemperature=3
    P.RequestCounters,  # ReadCounters=4
    P.RequestReadSettings,  # ReadSettings=5
    P.RequestConfig,  # WriteSettings=6
    P.RequestBalancePower,  # ReadBalancePower=7
    RequestTiming,  # Timing=8
    P.RequestBalanceCurrentCounter,  # ReadBalanceCurrentCounter=9
    P.RequestResetBalanceCurrentCounter,  # ResetBalanceCurrentCounter=10
    P.RequestBalanceLevel,  # WriteBalanceLevel=11
    P.RequestWritePIDconfig,  # 12
    P.RequestReadPIDconfig,  # 13
]

replyClass = [
    P.ReplyResetCounters,  # ResetCounters=0
    ReplyVoltages,  # ReadVoltages=1
    P.ReplyIdentify,  # Identify=2
    ReplyTemperature,  # ReadTemperature=3
    P.ReplyCounters,  # ReadCounters=4
    ReplyReadSettings,  # ReadSettings=5
    P.ReplyConfig,  # WriteSettings=6
    P.ReplyBalancePower,  # ReadBalancePower=7
    ReplyTiming,  # Timing=8
    P.ReplyBalanceCurrentCounter,  # ReadBalanceCurrentCounter=9
    P.ReplyResetBalanceCurrentCounter,  # ResetBalanceCurrentCounter=10
    P.ReplyBalanceLevel,  # WriteBalanceLevel=11
    P.ReplyWritePIDconfig,  # 12
    P.ReplyReadPIDconfig,  # 13
]



from __future__ import annotations  # noqa: D100

from moat.lib.micro import const

# bitfields

B_STREAM = const(1)
B_ERROR = const(2)
B_WARNING = const(3)  # external warning (no single integer allowed)
B_WARNING_INTERNAL = const(7)  # internal warning (may contain a naked int)

B_FLAGSTR = " SEW"

# errors

E_UNSPEC = const(-1)
E_NO_STREAM = const(-2)
E_CANCEL = const(-3)
E_NO_CMDS = const(-4)
E_SKIP = const(-5)
E_MUST_STREAM = const(-6)
E_ERROR = const(-7)
E_NO_CMD = const(-11)

# Stream states (separate for in/out)

S_END = const(3)  # terminal Stream=False message has been sent/received
S_NEW = const(4)  # No incoming message yet
S_ON = const(5)  # we're streaming (seen/sent first message)
S_OFF = const(6)  # in: we don't want streaming and signalled NO

# Stream directions

SD_NONE = const(0)
SD_IN = const(1)
SD_OUT = const(2)
SD_BOTH = const(3)

__all__ = [  # noqa:RUF022
    # bitfields
    "B_STREAM",
    "B_ERROR",
    "B_WARNING",
    "B_WARNING_INTERNAL",
    "B_FLAGSTR",
    # errors
    "E_UNSPEC",
    "E_NO_STREAM",
    "E_CANCEL",
    "E_NO_CMDS",
    "E_SKIP",
    "E_MUST_STREAM",
    "E_ERROR",
    "E_NO_CMD",
    # stream states
    "S_END",
    "S_NEW",
    "S_ON",
    "S_OFF",
    # stream directions
    "SD_NONE",
    "SD_IN",
    "SD_OUT",
    "SD_BOTH",
]



"""
System access app
"""

from __future__ import annotations

import time

from moat.util import enc_part, get_part, set_part
from moat.lib.path import Path
from moat.lib.proxy import Proxy, drop_proxy
from moat.lib.rpc import BaseCmd
from moat.micro.util import TEST_MAGIC

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from collections.abc import Callable


class Cmd(BaseCmd):
    """
    Generic system specific commands
    """

    doc = dict(_c=dict(_d="System level commands"))

    def __init__(self, cfg):
        super().__init__(cfg)
        self.cache = {}

    doc_test = dict(_d="enc test str", _r="str:fancy string")

    async def cmd_test(self):
        """
        Returns a test string: r CR n LF - NUL c ^C e ESC !

        Use this to verify that nothing mangles anything.
        """
        return TEST_MAGIC

    doc_unproxy = dict(_d="drop proxy", _0="Proxy")

    async def cmd_unproxy(self, p):
        """
        Tell the client to forget about a proxy.

        @p accepts either the proxy's name, or the proxied object.
        """
        if p == "" or p == "-" or p[0] == "_":
            raise RuntimeError("cannot be deleted")
        drop_proxy(p)

    doc_eval = dict(
        _d="eval",
        _0="Proxy|str|list",
        r="str|bool:cache result",
        a="list:fn args",
        k="dict:fn kwargs",
    )

    async def cmd_eval(self, x, r: str | bool = False, a=None, k=None):
        """
        Debugging/Introspection/Evaluation.

        @x can be
        * a string: evaluated, context is the eval cahce
        * a list: descend into an object.
          the first item must be a proxy (object reference)
          or a string (eval cache lookup)
        * an object (possibly proxied): left as-is

        Then if a or k is not None, assume that the result is a function
        and call it.

        If @r is  a string, the result is stored in the eval cache
        under that name. A list is interpreted as an accessor:
        ``x=42, r=('a','b','c')`` assigns ``cache['a'].b.c=42``.
        Nothing is returned in these cases.

        If @r is False *or* if the result is a dict or array, it is
        returned as a two-element list of (dict/list of simple members;
        list of indices of complex members). For objects, a third element
        contains the object type's name.
        If @r is True, the result's ``repr`` is returned.
        Otherwise (@r is ``None``), a proxy is returned.
        """
        if not self.cache:
            self.cache["self"] = self
            self.cache["root"] = self.root

        if isinstance(x, str):
            try:
                res = eval(x, self.cache)
            except SyntaxError:
                exec(x, self.cache)  # noqa:S102
                res = None
        elif isinstance(x, (tuple, list, Path)):
            res = x[0]
            if isinstance(res, str):
                res = self.cache[res]
            res = get_part(res, x[1:])
        else:
            res = x

        # call it?
        if a is not None or k is not None:
            res = cast("Callable", res)(*(a or ()), **(k or {}))
            if hasattr(res, "throw"):
                res = await res

        # store it?
        if isinstance(r, str):
            # None: drop from the cache
            if res is None:
                del self.cache[r]
            else:
                self.cache[r] = res
            return res is not None

        if isinstance(r, (list, tuple)):
            set_part(self.cache, r, res)
            return None

        if isinstance(res, (dict, list, tuple)):
            # dicts+lists always get encoded
            res = enc_part(res)
        elif not isinstance(res, (int, float, Proxy)):
            if r is True:
                return repr(res)
            if r is False:
                try:
                    rd = res.__dict__
                except AttributeError:
                    pass
                else:
                    res = enc_part(rd, res.__class__.__name__)
        return res

    doc_ping = dict(_d="Reply test", m="any:Return data", _r=dict(m="any:Return data"))

    async def cmd_ping(self, m=None):
        """
        Echo @m.

        This is for humans and testing. Don't use it for automated keepalive,
        that's what the "external" function of the watchdog is good for.
        """
        return {"m": m}

    doc_time = dict(
        _d="Get current satellite time(s)",
        _r=dict(t="int:seconds", ms="int:milliseconds", us="int:microseconds"),
    )

    async def cmd_time(self):
        """
        Return the current time.
        """
        f_ms = getattr(time, "ticks_ms", None)
        f_us = getattr(time, "ticks_us", None)
        return {
            "t": time.time(),
            "ms": f_ms() if callable(f_ms) else int(time.time() * 1000),
            "us": f_us() if callable(f_us) else int(time.time() * 1000000),
        }



"""
Basic message block
"""

from __future__ import annotations

from moat.util import ExpectedError, outcome, push_kw
from moat.lib.codec.errors import SilentRemoteError
from moat.lib.micro import (
    CancelledError,
    Event,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Queue,
    is_async,
    log,
    log_exc,
    shield,
)
from moat.lib.path import Path

from .const import (
    B_ERROR,
    B_STREAM,
    B_WARNING,
    B_WARNING_INTERNAL,
    E_CANCEL,
    E_ERROR,
    E_NO_STREAM,
    E_SKIP,
    S_END,
    S_NEW,
    S_OFF,
    S_ON,
    SD_BOTH,
    SD_IN,
    SD_NONE,
    SD_OUT,
)
from .errors import Flow, NoStream, StreamError, WantsStream

from typing import TYPE_CHECKING, cast, overload

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from moat.lib.path import PathElem

    from .base import OptDict

    from collections.abc import Callable, ItemsView, Iterator, KeysView, Sequence, ValuesView
    from typing import Any, Literal, Self


__all__ = ["Msg", "MsgResult"]


class MsgResult(Iterable):
    """
    This class encapsulates the result of a message, which is
    simultaneously a list and a dict. Both are read-only.

    You can access mutable versions with `args` and `kw`.
    """

    _a: Sequence
    _kw: dict

    def __init__(self, a: list, kw: dict):
        self._a = a
        self._kw = kw

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._a} {self._kw}>"

    @property
    def args(self) -> Sequence:
        "Retrieve the argument list."
        return self._a

    @property
    def args_l(self) -> MutableSequence:  # pyright:ignore
        """
        Retrieve the argument list, in a mutable form.
        """
        if not isinstance(self._a, MutableSequence):
            self._a = list(self._a)
        return self._a

    @property
    def kw(self) -> MutableMapping[str, Any]:
        "Retrieve the keywords."
        return self._kw

    @overload
    def to_list(self, dict_only: Literal[False]) -> list[Any]: ...

    @overload
    def to_list(self, dict_only: None) -> list[Any]: ...

    @overload
    def to_list(self, dict_only: Literal[True]) -> list[Any] | dict: ...

    def to_list(self, dict_only: bool | None = True) -> list[Any] | Mapping[str, Any]:
        """
        Returns a list of positional arguments.
        May end with a dict of keyword arguments.

        Args:
            dict_only:
                True (default):
                    If the positional args are empty but the keywords are
                    not, returns a dict, otherwise a list.
                False:
                    Always return the positional arguments followed by a
                    keyword mapping.
                None:
                    Append an empty keyword mapping only when required for
                    disambiguation.
        """
        if dict_only and self._kw and not self._a:
            return self._kw
        a = list(self._a)
        if dict_only is False:
            a.append(self._kw)
        else:
            push_kw(a, self._kw)
        return a

    def __len__(self) -> int:
        """
        Returns the number of positional values.

        :meta public:
        """
        return len(self._a)

    def __bool__(self) -> bool:
        """
        Always `True`.

        :meta public:
        """
        return True

    def __getitem__(self, k: int | str) -> Any:
        """
        Get an item. If the key is numeric, retrieve from the argument
        list, else from the keywords.

        :meta public:
        """
        if isinstance(k, (int, slice)):
            return self._a[k]
        return self._kw[k]

    def get(self, k: int | str, default=None, nulled=False) -> Any:
        """
        Get an item. Like :py:meth:`__getitem__` but returns a default (None) instead of
        raising `KeyError` / `IndexError`.
        """
        if isinstance(k, int):
            try:
                res = self._a[k]
            except IndexError:
                return default
        else:
            try:
                res = self._kw[k]
            except KeyError:
                return default

        if nulled and res is None:
            return default
        return res

    def __contains__(self, k: int | str) -> bool:
        """
        Test whether the given key is a valid positional index (if it's an
        integer) or an existing keyword (if a string).

        Use :py:attr:`kw` if you need to access numeric keywords.

        :meta public:
        """
        if isinstance(k, int):
            return 0 <= k < len(self._a)
        return k in self._kw

    def __iter__(self) -> Iterator:
        """
        Returns an iterator over the positional values.

        :meta public:
        """
        return iter(self._a)

    def keys(self) -> KeysView:
        """
        Returns an iterator over the keyword names.

        :meta public:
        """
        "Returns an iterator over the dict's keys."
        return self._kw.keys()

    def values(self) -> ValuesView:
        """
        Returns an iterator over the keyword values.

        :meta public:
        """
        "Returns an iterator over the dict's values."
        return self._kw.values()

    def items(self) -> ItemsView:
        """
        Returns an iterator over the keyword items.

        :meta public:
        """
        return self._kw.items()


_link_id = 0


class MsgLink:
    """
    The "other side" of a message.

    Message links are bidirectional tunnels. :meth:`ml_send` on one side delivers
    to the :meth:`ml_recv` method of the other.

    This class is the base implementation, which simply forwards data from
    a message to its sibling.
    """

    _remote: MsgLink | None = None
    _end: bool = False

    def __init__(self):
        """
        Link setup. By default does nothing.

        You need to call :meth:`set_remote` before this is usable.
        """
        global _link_id
        _link_id += 1
        self.link_id = _link_id

    async def ml_recv(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """Message Link Receive

        Called from the other side with whatever data.

        Override this.
        """
        raise NotImplementedError

    async def ml_send(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """Message Link Send

        This method forwards data to the other side.

        Don't override this.
        """
        if self._remote is None:
            raise EOFError
        try:
            await self._remote.ml_recv(a, kw, flags)
        except BaseException:
            await self.kill()
            raise
        else:
            if not flags & B_STREAM:
                self.set_end()

    async def ml_send_error(self, exc):
        """
        Send an exception.
        """
        if self.end_here:
            if isinstance(exc, Exception):
                log("Err after end: %r %r", self, exc, err=exc)
            return
        try:
            # send the error directly
            await self.ml_send((exc,), None, B_ERROR)
        except Exception as exc:
            try:
                # that failed? send the error name and arguments
                await self.ml_send((exc.__class__.__name__,) + exc.args, None, B_ERROR)
            except Exception:
                try:
                    # that failed too? send just the error name
                    await self.ml_send((exc.__class__.__name__,), None, B_ERROR)
                except Exception:
                    try:
                        # oh well, just send a naked error indication
                        await self.ml_send([E_ERROR], None, B_ERROR)
                    except Exception as ex:
                        log("Unable to send: %r %r / %r", self, exc, ex)
                        pass

    @property
    def end_both(self) -> bool:
        if self._remote and not self._remote.end_here:
            return False
        return self._end

    @property
    def end_here(self) -> bool:
        return self._end

    @property
    def end_there(self) -> bool:
        if self._remote is None:
            return True
        return self._remote.end_here

    @property
    def remote(self):
        return self._remote

    def stream_detach(self):
        """
        Called when this stream is done.

        Override this method to clean up any related data.

        This method must be idempotent.
        """
        pass

    def set_end(self):
        "The send side of this stream has ended."
        self._end = True
        if self.end_both:
            if self._remote is not None:
                self._remote.stream_detach()
            self.stream_detach()
            self._remote = None

    async def kill(self):
        """
        No further communication is possible.

        This must only be called if e.g. a link is down.
        It tries to deliver cancel messages to both sides.
        """
        rem = self._remote
        if self._end:
            pass
        else:
            rs = self if rem is None else rem
            rs.set_end()
            try:
                with shield():
                    await rs.ml_recv([E_CANCEL], None, B_ERROR)
            except Exception:  # noqa:S110
                pass

    def set_remote(self, remote: MsgLink):
        """
        Set (or change) my remote side.

        Args:
            remote: the new MsgLink on the remote side.

        The old remote, if any, will be :meth:`kill`-ed.
        """
        rem = self._remote
        if rem is not None:
            rem._remote = None  # noqa: SLF001
            rem.set_end()
        self._remote = remote

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}:L{self.link_id} "
            f"r{'=L' + str(self._remote.link_id) if self._remote else '-'}>"
        )


class Msg(MsgLink, MsgResult):
    """
    Message encapsulation and data streaming.
    """

    # The multiple inheritance problem WRT µPy is resolved below.

    _cmd: Path | None = None
    _a: Sequence
    _kw: dict
    # also in MsgResult

    _stream_in: int = S_NEW
    _stream_out: int = S_NEW

    _dir: int = SD_NONE

    _msg: outcome.Outcome | None = None
    _msg2: outcome.Outcome | None = None
    _msg_in: Event
    _recv_q: Queue | None = None
    _recv_qlen: int = 20
    _recv_skip: bool = False

    _fli: int | None = None
    _flo_evt: Event | None = None
    warnings: list

    _loaded: bool = False

    def __init__(self):
        """
        Set up the message.
        """
        super().__init__()
        self._msg_in = Event()
        self.warnings = []  # TODO

    @property
    def cmd(self) -> Path | None:
        "Retrieve the command."
        return self._cmd

    @property
    def rcmd(self) -> list[PathElem]:
        """
        Retrieve a reversed command
        """
        assert self.cmd is not None
        res = list(self.cmd)
        res.reverse()
        return cast("list[PathElem]", res)

    @classmethod
    def Call(cls, cmd: Path, a: list, kw: dict, flags: int = 0) -> Self:
        """Constructor for a possibly-remote function call."""
        if isinstance(cmd, str):
            if ":" in cmd or "." in cmd:
                raise ValueError("Wrap command paths in a `P()` call")
            cmd = Path.build((cmd,))
        s = cls()
        s._cmd = cmd
        s._a = a
        s._kw = kw
        if flags & B_STREAM:
            s._stream_in = S_ON
        return s

    @property
    def remote(self) -> MsgLink | None:  # noqa: D102
        return self._remote

    def replace_with(self, link: MsgLink) -> None:
        """
        Tell my own remote to point to @link instead.
        """
        if (rem := self._remote) is None:
            # we are a straight command handler and don't yet have a remote.
            link.set_remote(self)
            self._remote = link
            return

        rem.set_remote(link)  # this kills self
        link.set_remote(rem)

    async def kill(self, new: bool = False) -> None:
        """No further communication may happen on this message.

        If @new is set, this not being a "new" stream will raise a runtime
        exception.
        """
        try:
            if new:
                if self._stream_in != S_NEW:
                    raise RuntimeError("incoming already started")
                if self._stream_out != S_NEW:
                    raise RuntimeError("outgoing already started")
        finally:
            self._stream_in = S_END
            self._stream_out = S_END
            if self._msg_in is not None:
                self._msg_in.set()
            await super().kill()

    async def ml_send(
        self, a: Sequence, kw: OptDict, flags: int, initial: bool | None = None
    ) -> None:
        """
        Sender of data to the other side.
        """
        if self._stream_out == S_END:
            return
        if not flags & B_STREAM:
            self._stream_out = S_END
        else:
            if self._stream_out == S_NEW and not flags & B_ERROR:
                self._stream_out = S_ON
            if initial is False and self._stream_in == S_NEW:
                self._stream_in = S_ON
        await super().ml_send(a, kw, flags)

    async def ml_recv(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """
        Receiver for data from the other side.
        """
        # if S_END, no message may be exchanged
        # else if Stream bit is False, stop streaming if it is on, go to S_END: out of band
        # else if Error bit is True: flow / warning
        # else if S_NEW: go to S_ON: out-of-band
        # else: streamed data

        if self._stream_in == S_END:
            # This is a late-delivered incoming-stream-terminating error.
            if not flags & B_ERROR or len(a) != 1 or a[0] != E_CANCEL:
                # Don't log if cancelled
                log("LATE? L%d %r %r %d", self.link_id, a, kw, flags)

        elif not flags & B_STREAM:
            self._set_msg(a, kw, flags)
            self._stream_in = S_END
            if self._stream_out == S_ON:
                self._stream_out = S_OFF
            if self._recv_q is not None:
                self._recv_q.close_sender()

        elif flags & B_ERROR:
            # Warning.
            if len(a) == 1 and not kw and flags == B_WARNING_INTERNAL:
                pass
            elif kw or (a and isinstance(a[-1], Mapping)):
                if not hasattr(a, "append"):
                    a = list(a)
                a = cast(list, a)
                a.append(kw)
            exc = StreamError(a)
            if isinstance(exc, Flow):
                if self._flo_evt is None:
                    self._flo = exc.n
                    self._flo_evt = Event()
                else:
                    if self._flo == 0:
                        self._flo_evt.set()
                        self._flo_evt = Event()
                    self._flo += exc.n
                # otherwise ignore
            elif self._stream_in == S_ON and self._recv_q is not None:
                await self._recv_q.put_error(exc)
            else:
                self.warnings.append(exc)

        elif self._stream_in == S_NEW:
            self._set_msg(a, kw, flags)
            self._stream_in = S_ON

        elif self._recv_q is not None:
            await self._recv_q.put((a, kw))

        else:
            log("Unwanted stream: %r/%r/%d", a, kw, flags)
            if self._stream_in == S_ON:
                self._stream_in = S_OFF
                if self._stream_out != S_END:
                    await self.ml_send([E_NO_STREAM], None, B_ERROR)
                    self._stream_out = S_END

        await self._ended()

    async def send(self, *a, **kw) -> None:
        """
        Send a streamed message.
        """
        if not self._dir & SD_OUT:
            raise RuntimeError("This stream is read only")
        if self._stream_out != S_ON:
            log("NoStream:%d", self._stream_out)
            raise NoStream
        await self._skipped()
        await self.ml_send(a, kw, B_STREAM)

    async def warn(self, *a, **kw) -> None:
        """
        Send a warning on this stream (single integers are masked).
        """
        await self.ml_send(a, kw, B_WARNING)

    async def warn_int(self, *a, **kw) -> None:
        """
        Send an internal warning (single integers are not masked).
        """
        await self.ml_send(a, kw, B_WARNING_INTERNAL)

    async def error(self, *a, **kw) -> None:
        """
        Reply / end the stream with an error.
        """
        await self.ml_send(a, kw, B_ERROR)

    def _set_msg(self, a: Sequence, kw: OptDict, flags: int) -> None:
        """
        A message has arrived on this stream. Store and set an event.
        """
        if flags & B_ERROR:
            msg = outcome.Error(StreamError(a))
        else:
            if kw is None:
                kw = {}
            msg = outcome.Value((a, kw))

        if self._msg is None:
            self._msg = msg
            self._msg_in.set()
        elif self._msg2 is None:
            self._msg2 = msg
        else:
            raise RuntimeError("Msg Collision?")

    async def _ended(self) -> None:
        """
        If message processing is finished, finalize processing this
        message. Otherwise do nothing.
        """
        if self._stream_in != S_END:
            return
        if self._stream_out != S_END:
            return
        await self.kill()

    # Stream starters

    async def prep_stream(self, flag: int) -> None:
        """Sets up streaming as per SD_* flags.

        Sends an E_NO_STREAM warning if there's no streaming but queued data.
        """
        self._dir = flag

        if flag & SD_IN:
            if self._recv_q is None:
                self._recv_q = Queue(self._recv_qlen)
        else:
            q, self._recv_q = self._recv_q, None
            if q is not None and q.qsize() and self._stream_in == S_ON:
                self._stream_in = S_OFF
                await self.warn_int(E_NO_STREAM)
            # whatever has been received will be discarded

        self._stream_out = S_ON if flag & SD_OUT else S_OFF

    async def no_stream(self) -> None:
        """Mark as neither send or receive streaming."""
        if self._stream_in == S_ON:
            if self._stream_out != S_END:
                await self.error(E_NO_STREAM)
            raise WantsStream
        self._recv_q = None
        self._dir = 0
        # TODO

    # Stream reply helpers

    def stream_in(self, *a, **kw) -> AbstractAsyncContextManager[Msg]:
        """
        Reply to a stream, read-only.
        """
        return self._stream(a, kw, SD_IN)

    def stream_out(self, *a, **kw) -> AbstractAsyncContextManager[Msg]:
        """
        Reply to a stream, write-only.
        """
        return self._stream(a, kw, SD_OUT)

    def stream(self, *a, **kw) -> AbstractAsyncContextManager[Msg]:
        """
        Reply to a stream, bidirectional.
        """
        return self._stream(a, kw, SD_BOTH)

    def _stream_call(self, flag: int) -> AbstractAsyncContextManager[Msg]:
        """
        Stream startup on the sending side.  Don't call this.
        """
        return self._stream((), None, flag, initial=True)

    @property
    def can_stream(self) -> bool:
        """check whether this is a streaming command"""
        if self._stream_in == S_END or self._stream_out == S_END:
            return False
        if self._stream_in != S_NEW or self._stream_out != S_NEW:
            return True
        if (rem := self.remote) is None:
            return False
        if not isinstance(rem, Msg):
            return False
        try:
            if rem._stream_in != S_NEW or rem._stream_out != S_NEW:  # noqa: SLF001
                return True
        except AttributeError:
            pass
        return False

    async def call_simple(self, cmd: Callable) -> None:
        """Handle a non-streamed call endpoint.

        Args:
            cmd: a callable that takes whichever arguments the message contains.
                Typically a ``cmd_*`` method.
        """
        try:
            res = cmd(*self._a, **self._kw)
            if is_async(res):
                res = await res
        except Exception as exc:
            if self._remote is None:
                raise
            if not isinstance(exc, SilentRemoteError) and not isinstance(exc, ExpectedError):
                if exc.args != ("schema",):
                    log_exc(exc, "Command Error %r", self)
            await self.ml_send_error(exc)
        except BaseException as exc:
            if self._remote is None:
                raise
            log_exc(exc, "Command Error %r", self)
            await self.ml_send_error(exc)
            raise
        else:
            if isinstance(res, Msg):
                await self.result(*res.args, **res.kw)  # pyright:ignore
            elif isinstance(res, MutableMapping):
                await self.result(**res)  # pyright:ignore
            else:
                await self.result(res)

    def ensure_remote(self):
        """
        A context mamager that adds a remote side to an existing message.
        """

        return _EnsureRemote(self)

    async def call_stream(self, cmd: Callable) -> None:
        """Handle a streamed call endpoint.

        Args:
            cmd: an async callable, typically a ``stream_*`` method.
        """
        # If this message is direct and doesn't yet have a counterpart,
        # create one and re-do the call on that.

        if self._remote is None:
            async with self.ensure_remote() as m:
                return await m.call_stream(cmd)
        try:
            await cmd(self)
        except Exception as exc:
            if not isinstance(exc, ExpectedError):
                log_exc(exc, "Stream Error %r", self)
            await self.ml_send_error(exc)
        except BaseException as exc:
            await self.ml_send_error(exc)
            raise
        else:
            if self._stream_out != S_END:
                await self.result()

    def _stream(self, a: Sequence, kw: OptDict, flag: int, initial: bool = False):
        return _Stream(self, a, kw, flag, initial=initial)

    async def result(self, *a, **kw) -> None:
        """
        Send (or set) the result.
        """
        if self._remote is None:
            if self._msg is not None:
                if kw:
                    raise RuntimeError("Dup call", kw)
                if a and (len(a) > 1 or a[0] is not None):  # pyright:ignore
                    raise RuntimeError("Dup call", a)

            self._msg = outcome.Value((a, kw))
            self._msg_in.set()
            return

        await self.ml_send(a, kw, 0)

    async def wait_replied(self, preload: bool = False) -> None:
        """
        Wait for a (non-streamed) reply.
        """
        loaded, self._loaded = self._loaded, preload
        if loaded:
            return

        if self._msg is None:
            if self._stream_in == S_END:
                raise NoStream
            await self._msg_in.wait()
            self._msg_in = Event()
        msg = self._msg
        self._msg = self._msg2
        self._msg2 = None
        if msg is None:
            raise EOFError
        if isinstance(msg, outcome.Error):
            msg.unwrap()
            return
        msg = cast("outcome.Value[tuple[list, dict]]", msg)
        self._a, self._kw = msg.unwrap()

    def __aiter__(self) -> Self:
        if not self._dir & SD_IN:
            raise RuntimeError("This stream is write only")
        return self

    async def _skipped(self):
        """
        Test whether incoming data could not be delivered due to the
        receive queue getting full.
        """
        if self._recv_q is not None and self._recv_skip and self.stream_out != S_END:
            await self.warn_int(E_SKIP)
            self._recv_skip = False

    async def _qsize(self) -> None:
        # Incoming message queue handling strategy:
        # - read without flow control until the queue is half full
        assert self._recv_q is not None  # to shut up pyright

        if self._fli is None:
            if self._recv_q.qsize() >= self._recv_qlen // 2:
                self._fli = 0
                # - send a message announcing 1/4 of the queue space
                await self.warn_int(self._recv_qlen // 4)

        # - then, whenever the queue is at most 1/4 full *and* qlen/2 messages
        #   have been processed (which will happen because the queue was
        #   half-full when we started), announce that space
        elif self._recv_q.qsize() <= self._recv_qlen // 4 and self._fli > self._recv_qlen // 2:
            m = self._recv_qlen // 2 + 1
            self._fli -= m
            await self.warn_int(m)

        # - additionally, if the max queue is < 10
        #   we send a bit more aggressively, to reduce lag
        else:
            self._fli += 1
            if self._recv_qlen < 10 and self._fli >= self._recv_qlen // 4:
                m, self._fli = self._fli, 0
                await self.warn_int(m)

    async def __anext__(self) -> MsgResult:
        if self._recv_q is None:
            raise StopAsyncIteration
        elif isinstance(self._recv_q, Exception):
            exc, self._recv_q = self._recv_q, None
            raise exc
        await self._skipped()

        try:
            res = await self._recv_q.get()
        except EOFError:
            raise StopAsyncIteration from None

        await self._qsize()
        return MsgResult(*res)

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}:L{self.link_id} "
            f"r{'=L' + str(self._remote.link_id) if self._remote else '-'}: "
            f" {' ' + str(self._cmd) if self._cmd else ''} {self._a} {self._kw}>"
        )


class _Stream:
    def __init__(self, slf, a: Sequence, kw: OptDict, flag: int, initial: bool = False):
        self.slf = slf
        self.a = a
        self.kw = kw
        self.flag = flag
        self.initial = initial

    async def __aenter__(self):
        slf = self.slf
        if slf._stream_out != S_NEW:  # noqa: SLF001
            raise RuntimeError(
                "Simple command" if slf._stream_out == S_END else "Stream-out already set",  # noqa: SLF001
            )

        # stream-in depends on what the remote side sent
        await slf.prep_stream(self.flag)

        if slf._recv_qlen < 10:  # noqa: SLF001
            slf._fli = 0  # noqa: SLF001
            await slf.warn_int(slf._recv_qlen)  # noqa: SLF001

        if self.initial:
            await slf.wait_replied()
        else:
            await slf.ml_send(self.a, self.kw, B_STREAM, self.initial)
        return slf

    async def _close(self):
        # This code is running inside the handler, which will process the error
        # case. Thus we don't need error handling here.

        slf = self.slf
        if slf._stream_out != S_END:  # noqa: SLF001
            await slf.ml_send([None], {}, 0)

        await slf.wait_replied()
        if slf._stream_in != S_END:  # noqa: SLF001
            raise RuntimeError("Stream not ended")

    async def __aexit__(self, c, e, t):
        try:
            slf = self.slf
            if e is not None and slf._stream_out != S_END:  # noqa: SLF001
                await slf.ml_send_error(e)
            try:
                await self._close()
            except CancelledError:
                pass
        finally:
            if e is not None:
                raise e


class _EnsureRemote:
    def __init__(self, slf):
        self.slf = slf

    async def __aenter__(self):
        slf = self.slf
        if (m := slf._remote) is None:  # noqa: SLF001
            m = Msg.Call(slf._cmd, slf._a, slf._kw)  # noqa: SLF001
            slf._cmd, slf._a, slf._kw = None, (), {}  # noqa: SLF001
            m.set_remote(slf)
            slf.set_remote(m)
        self.m = m
        return m

    async def __aexit__(self, *err):
        with shield():
            await self.m.kill()
            await self.slf.kill()



"""
Support code to connect to a TCP server.
"""

from __future__ import annotations

from moat.lib.micro import Event, L, TaskGroup, run_server
from moat.lib.stream import SingleAIOBuf

from .util import BaseConnIter

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from typing import Never


class TcpIter(BaseConnIter):
    """
    A connection iterator for TCP sockets
    """

    def __init__(self, host, port):
        super().__init__()
        self.host = host
        self.port = port

    async def accept(self) -> Never:  # noqa:D102
        async with TaskGroup() as tgx:
            evt = Event()

            async def rdy(evt):
                await evt.wait()
                if L:
                    self.set_ready()

            await tgx.spawn(rdy, evt)

            await run_server(self._handle, self.host, self.port, evt=evt)

    async def _handle(self, conn):
        await self.add_conn(SingleAIOBuf(conn))



"""
Fake RTC backend for testing.

Provides an RTC class that stores data in memory instead of hardware RTC.
"""

from __future__ import annotations

from moat.util import NotGiven

from . import RTCBase


class RTC(RTCBase):
    """
    Fake RTC backend for testing.

    Stores data in memory. Behaves like mem.py but without hardware.
    """

    is_FS = False
    is_ASYNC = False

    def __init__(self, cfg=None):
        if cfg is None:
            cfg = {}
        super().__init__(cfg)
        self._d = {}

    def reload(self):
        """Reload is a no-op for the fake backend."""
        pass

    async def get(self, name: str):
        """Get a value."""
        if name not in self._d:
            raise KeyError(name)
        return self._d[name]

    async def set(self, name: str, data) -> bool:
        """Set a value."""
        if data is NotGiven:
            self._d.pop(name, None)
        else:
            self._d[name] = data
        return True

    async def keys(self) -> set[str]:
        """Get keys of stored data."""
        return set(self._d.keys())
